#include "./inc/bi_fold_stack.h"
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include <algorithm>
#include <stack>

using namespace std;

bi_fold_stack::bi_fold_stack(string input_sequence, string input_SHAPE_file, 
	int input_frequency, int input_min_stack_length, int input_min_stack_hydro_bonds)	{
	sequence = input_sequence;
	SHAPE_file = input_SHAPE_file;
	shape_reactivity SHAPE_constraint(sequence, SHAPE_file);
	reactivity = SHAPE_constraint.reactivity;
	frequency = input_frequency;
	structure_1 = string(sequence.length(), 'x');
	structure_2 = string(sequence.length(), 'x');
	/* predict stacks based on input criterions */
	min_stack_length = input_min_stack_length;
	min_stack_hydro_bonds = input_min_stack_hydro_bonds;
	
	int i, j;
	stack_hash = vector< vector<int> > (sequence.length());
	for(i = 0; i < sequence.length(); ++ i)	{
		stack_hash[i] = vector<int> (sequence.length());
		for(j = 0; j < sequence.length(); ++ j)	{
			stack_hash[i][j] = -1;
		}
	}
		
	predict_stacks(min_stack_length, min_stack_hydro_bonds);
	sort_stacks();
	build_stacks();
	/* define vector size for DP matrices */
	unconstrained_matrix = vector<folding_matrix> (stacks.size());
	unconstrained_multi_matrix = vector< vector<folding_matrix> > (stacks.size());
	constrained_matrix = vector< vector<folding_matrix> > (stacks.size());
	chainloop_matrix_iloop = vector< vector< vector<folding_matrix> > >  (stacks.size());
	chainloop_matrix_jloop = vector< vector< vector<folding_matrix> > >  (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		unconstrained_multi_matrix[i] = vector<folding_matrix> (stacks.size());
		constrained_matrix[i] = vector<folding_matrix> (stacks.size());
		chainloop_matrix_iloop[i] = vector< vector<folding_matrix> > (stacks.size());
		chainloop_matrix_jloop[i] = vector< vector<folding_matrix> > (stacks.size());
		for(j = 0; j < stacks.size(); ++ j)	{
			chainloop_matrix_iloop[i][j] = vector<folding_matrix> (stacks.size());
			chainloop_matrix_jloop[i][j] = vector<folding_matrix> (stacks.size());
		}
	}
	return;
		
}

bi_fold_stack::~bi_fold_stack(void)	{
	return;
}

int bi_fold_stack::predict_stacks(int min_length, int min_hydro_bonds)	{
	int i, j, k;
	
	energy_parameters energy_model(sequence);
	
	vector< vector<int> > sampled_pairs(sequence.length());
	for(i = 0; i < sequence.length(); ++ i)	{
		sampled_pairs[i] = vector<int> (sequence.length());
		for(j = 0; j < sequence.length(); ++ j)	{
			sampled_pairs[i][j] = 0;
		}
	}
	
	for(i = 0; i < sequence.length() - 4; ++ i)	{
		for(j = i + 4; j < sequence.length(); ++ j)	{
			int stack_hydro_bonds = num_hydro_bonds(sequence[i], sequence[j]);
			if(stack_hydro_bonds < 0 || sampled_pairs[i][j] == 1) continue;
			k = 1;
			/* find continuous stackings */
			while(j - i - 2 * k - 1 >= 3)	{
				int p_hydro = num_hydro_bonds(sequence[i + k], sequence[j - k]);
				if(p_hydro > 0)	stack_hydro_bonds += p_hydro;
				else	break;
				++ k;
			}
			/* removing terminal GU pairs */
			int l = 1;
			/*******************************
			while(l <= k)	{
				int p_hydro = num_hydro_bonds(sequence[i + k - l], sequence[j - k + l]);
				if(p_hydro == 1)	stack_hydro_bonds -= p_hydro;
				else	break;
				++ l;
			}
			*******************************/
			
			if(stack_hydro_bonds >= min_hydro_bonds && k - l + 1 >= min_length)	{
				stacks.resize(stacks.size() + 1);
				stacks[stacks.size() - 1] = vector<int> (4);
				stacks[stacks.size() - 1][0] = i;
				stacks[stacks.size() - 1][1] = j;
				stacks[stacks.size() - 1][2] = k - l + 1;
				int m;
				int energy = 0;
				for(m = k - l - 1; m >= 0; -- m)	{
					energy += energy_model.FE_stack(i + m, j - m);
				}
				stacks[stacks.size() - 1][3] = energy;
				for(m = 0; m < k - l + 1; ++ m)	{
					sampled_pairs[i + m][j - m] = 1;
				} 
				
			}
		}
	}
	/* add the pseudo stack to cover the entire structure */
	stacks.resize(stacks.size() + 1);
	stacks[stacks.size() - 1] = vector<int> (4);
	stacks[stacks.size() - 1][0] = 0;
	stacks[stacks.size() - 1][1] = sequence.length() - 1;
	stacks[stacks.size() - 1][2] = 0;
	stacks[stacks.size() - 1][3] = 0;
	
	return stacks.size();
}

void bi_fold_stack::sort_stacks()	{
	
	int i, j;
	for(i = 0; i < stacks.size() - 1; ++ i)	{
		for(j = i + 1; j < stacks.size(); ++ j)	{
			if(stacks[i][1] > stacks[j][1] || (stacks[i][1] == stacks[j][1] && stacks[i][0] < stacks[j][0]))	{
				vector<int> temp(4);
				temp = stacks[i]; stacks[i] = stacks[j]; stacks[j] = temp;
			}
		}
	}
	
	/* construct the stack hash */
	for(i = 0; i < stacks.size(); ++ i)	{
		stack_hash[stacks[i][0]][stacks[i][1]] = i;
	}
	
	return;
	
}

int bi_fold_stack::num_hydro_bonds(char a, char b)	{
	
	a = tolower(a);
	b = tolower(b);
	
	if(a == 'a' && b == 'u' || a == 'u' && b == 'a')	{
		return 2;
	}	else if(a == 'c' && b == 'g' || a == 'g' && b == 'c')	{
		return 3;
	}	else if(a == 'g' && b == 'u' || a == 'u' && b == 'g')	{
		return 1;
	}	else	{
		return -LARGENUM;
	}
}

void bi_fold_stack::fill_unconstrained_matrix(void)	{
	
	
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	single_fold_stack free_fold(sequence, SHAPE_file, min_stack_length, min_stack_hydro_bonds);
	free_fold.fold_SHAPE();
	int i, j;
	for(i = 0; i < stacks.size(); ++ i)	{
		free_fold.clear_structure();
		free_fold.traceback_stack(i);
		compress_structure(0, free_fold.structure, unconstrained_matrix[i]);
		compress_structure(1, free_fold.structure, unconstrained_matrix[i]);
		unconstrained_matrix[i].str_1 = unconstrained_matrix[i].str_2 = free_fold.structure;
		
		unconstrained_matrix[i].free_energy_1 = free_fold.energy_matrix[i];
		SHAPE_constraint.adjust_unconstrained_energy(free_fold.structure, unconstrained_matrix[i].free_energy_1);
		unconstrained_matrix[i].free_energy_2 = unconstrained_matrix[i].free_energy_1;
		for(j = num_enclosing[i] - 1; j >= 0; -- j)	{
			int x = enclosing[i][j];
			
			free_fold.clear_structure();
			free_fold.traceback_multi(i, x);
			compress_structure(0, free_fold.structure, unconstrained_multi_matrix[i][x]);
			compress_structure(1, free_fold.structure, unconstrained_multi_matrix[i][x]);
			unconstrained_multi_matrix[i][x].str_1 = unconstrained_multi_matrix[i][x].str_2 = free_fold.structure;
			unconstrained_multi_matrix[i][x].free_energy_1 = free_fold.energy_matrix_multi[i][x];
			SHAPE_constraint.adjust_unconstrained_energy(free_fold.structure, unconstrained_multi_matrix[i][x].free_energy_1);
			unconstrained_multi_matrix[i][x].free_energy_2 = unconstrained_multi_matrix[i][x].free_energy_1;
		}
	}
	return;
	
}


void bi_fold_stack::compress_structure(bool which, string s, folding_matrix &m)	{
	
	int i = 0, j = sequence.length() - 1;
	
	/* find out boundaries */
	while(s[i] == 'x')	{
		++ i;
	}
	while(s[j] == 'x')	{
		-- j;
	}
	if(which == 0)	{
		m.str_begin_1 = i; 
		m.str_end_1 = j;
	}	else	{
		m.str_begin_2 = i; 
		m.str_end_2 = j;
	}
	
	if(which == 0)	{
		m.str_stacks_1.resize(0);
	}	else	{
		m.str_stacks_2.resize(0);
	}
	
	stack<int> hold_pairs;
	int k, left = -1, right = -1, prev_left = -1, prev_right = -1;
	for(k = i; k <= j; ++ k)	{
		if(s[k] == '(')	{
			hold_pairs.push(k);
		}	else if(s[k] == ')')	{
			left = hold_pairs.top();
			right = k;
			hold_pairs.pop();
			if((left != prev_left - 1 || right != prev_right + 1) 
				&& prev_left >= 0 && prev_right >= 0 && stack_hash[prev_left][prev_right] >= 0)	{
				if(which == 0)	{
					m.str_stacks_1.resize(m.str_stacks_1.size() + 1);
					m.str_stacks_1[m.str_stacks_1.size() - 1] = stack_hash[prev_left][prev_right];
				}	else	{
					m.str_stacks_2.resize(m.str_stacks_2.size() + 1);
					m.str_stacks_2[m.str_stacks_2.size() - 1] = stack_hash[prev_left][prev_right];
				}	
			}
			prev_left = left;
			prev_right = right;
		}
	}
	if(left >= 0 && right >= 0)	{
		if(which == 0)	{
			m.str_stacks_1.resize(m.str_stacks_1.size() + 1);
			m.str_stacks_1[m.str_stacks_1.size() - 1] = stack_hash[left][right];
		}	else	{
			m.str_stacks_2.resize(m.str_stacks_2.size() + 1);
			m.str_stacks_2[m.str_stacks_2.size() - 1] = stack_hash[left][right];
		}
	}
	return;
}

string bi_fold_stack::decompress_structure(bool which, folding_matrix m)	{
	
	int i;
	string d_str(sequence.length(), 'x');
	if(which == 0)	{
		if(m.str_begin_1 > m.str_end_1 || m.str_begin_1 < 0 || m.str_end_1 >= sequence.length())	{
			cout << "Error in decompress structure" << endl;
			return d_str;
		}
		d_str.replace(m.str_begin_1, m.str_end_1 - m.str_begin_1 + 1, string(m.str_end_1 - m.str_begin_1 + 1, '.'));
		for(i = 0; i < m.str_stacks_1.size(); ++ i)	{
			//cout << "@:	" << m.str_stacks_1[i] << endl;
			//cout << stacks[m.str_stacks_1[i]][0] << "	" << stacks[m.str_stacks_1[i]][1] << endl;
			append_stack(stacks[m.str_stacks_1[i]], d_str);
			//cout << "finish appending stack 1" << endl;
		}
	}	else	{
		if(m.str_begin_2 > m.str_end_2 || m.str_begin_2 < 0 || m.str_end_2 >= sequence.length())	{
			cout << "Error in decompress structure" << endl;
			return d_str;
		}
		d_str.replace(m.str_begin_2, m.str_end_2 - m.str_begin_2 + 1, string(m.str_end_2 - m.str_begin_2 + 1, '.'));
		for(i = 0; i < m.str_stacks_2.size(); ++ i)	{
			//cout << "&:	" << m.str_stacks_2[i] << endl;
			//cout << stacks[m.str_stacks_2[i]][0] << "	" << stacks[m.str_stacks_2[i]][1] << endl;
			append_stack(stacks[m.str_stacks_2[i]], d_str);
			//cout << "finish appending stack 2" << endl;
		}
	}
	//cout << d_str << endl;
	return d_str;
}

void bi_fold_stack::build_stacks(void)	{
	
	int i, j;
	
	/* building the internal relationship */
	num_internal = vector<int> (stacks.size());
	internal = vector< vector<int> > (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		num_internal[i] = 0;
		internal[i] = vector<int> (stacks.size());
	}
	
	for(i = stacks.size() - 1; i >= 0; -- i )	{
		for(j = i - 1; j >= 0; -- j)	{
			if(stacks[j][0] > stacks[i][0] + stacks[i][2] - 1 && stacks[j][1] < stacks[i][1] - stacks[i][2] + 1 &&
				stacks[j][0] - stacks[i][0] - stacks[i][2] <= 15 && 
				stacks[i][1] - stacks[i][2] - stacks[j][1] <= 15)	{
				internal[i][num_internal[i]] = j;
				++ num_internal[i];
			}
		}
	}
	
	/* building the enclosing relationship */
	num_enclosing = vector<int> (stacks.size());
	enclosing = vector< vector<int> > (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		num_enclosing[i] = 0;
		enclosing[i] = vector<int> (stacks.size());
	}
	
	for(i = stacks.size() - 1; i >= 0; -- i )	{
		for(j = i - 1; j >= 0; -- j)	{
			if(stacks[j][0] > stacks[i][0] + stacks[i][2] - 1 && stacks[j][1] < stacks[i][1] - stacks[i][2] + 1)	{
				enclosing[i][num_enclosing[i]] = j;
				++ num_enclosing[i];
			}
		}
	}
	
	/* building the FP relationship */
	num_FPset = vector<int> (stacks.size());
	FPset = vector< vector<int> > (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		num_FPset[i] = 0;
		FPset[i] = vector<int> (stacks.size());
	}
	for(i = stacks.size() - 1; i >= 0; -- i)	{
		int boundary = stacks[i][0] + stacks[i][2];
		for(j = i - 1; j >= 0; -- j)	{
			if(stacks[j][1] >= boundary && stacks[j][0] >= stacks[i][0] + stacks[i][2] && stacks[j][1] <= stacks[i][1] - stacks[i][2])	{
				if(stacks[j][0] > boundary)	{
					boundary = stacks[j][0];
				}
				FPset[i][num_FPset[i]] = j;
				++ num_FPset[i];
			}
		}
	}
	
	/* building the F relationship */
	num_Fset = vector<int> (stacks.size());
	Fset = vector< vector<int> > (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		num_Fset[i] = 0;
		Fset[i] = vector<int> (stacks.size());
	}
	for(i = stacks.size() - 1; i >= 0; -- i)	{
		int boundary = 0;
		for(j = i - 1; j >= 0; -- j)	{
			if(stacks[j][1] >= boundary && stacks[j][1] < stacks[i][0])	{
				if(stacks[j][0] > boundary)	{
					boundary = stacks[j][0];
				}
				Fset[i][num_Fset[i]] = j;
				++ num_Fset[i];
			}
		}
	}
	/*
	cout << "Printing all stacks" << endl;
	for(i = 0; i < stacks.size(); ++ i)	{
		cout << i << ":	" << stacks[i][0] << "	" << stacks[i][1] << "	" << stacks[i][2] << "	" << stacks[i][3] << endl;
	}
	cout << "End of printing all stacks" << endl;
	*/
	/*
	for(i = 0; i < stacks.size(); ++ i)	{
		cout << i << ":	";
		for(j = 0; j < num_Fset[i]; ++ j)	{
			cout << Fset[i][j] << "	";
		}
		cout << endl;
	}
	*/
	return;
}

bool bi_fold_stack::is_overlap(vector<int> stack_1, vector<int> stack_2)	{
	if(stack_1[0] > stack_2[1] || stack_1[1] < stack_2[0])	{
		return false;
	}	else {
		return true;
	}
}

void bi_fold_stack::append_stack(vector<int> stack, string &structure)	{
	string left(stack[2], '(');
	string right(stack[2], ')');
	structure.replace(stack[0], stack[2], left);
	structure.replace(stack[1] - stack[2] + 1, stack[2], right);
	return;
}

void bi_fold_stack::fold_unconstrained(int i, int j, folding_matrix &record_unconstrained)	{
	/* adding up the free energy */
	record_unconstrained.free_energy_1 = unconstrained_matrix[i].free_energy_1;
	record_unconstrained.free_energy_2 = unconstrained_matrix[j].free_energy_1;
	record_unconstrained.sum_free_energy = record_unconstrained.free_energy_1 + record_unconstrained.free_energy_2;
	/* record the structural constraint for 1st transcript */
	record_unconstrained.str_begin_1 = unconstrained_matrix[i].str_begin_1;
	record_unconstrained.str_end_1 = unconstrained_matrix[i].str_end_1;
	record_unconstrained.str_stacks_1 = unconstrained_matrix[i].str_stacks_1;
	record_unconstrained.str_1 = unconstrained_matrix[i].str_1;
	/* record the structural constraint for 2nd transcript */
	record_unconstrained.str_begin_2 = unconstrained_matrix[j].str_begin_1;
	record_unconstrained.str_end_2 = unconstrained_matrix[j].str_end_1;
	record_unconstrained.str_stacks_2 = unconstrained_matrix[j].str_stacks_1;
	record_unconstrained.str_2 = unconstrained_matrix[j].str_1;
	return;
}

void bi_fold_stack::fold_hairpin(int i, int j, folding_matrix &record_hairpin)	{
	
	energy_parameters energy_model(sequence);
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	
	string structure_new1(sequence.length(), 'x');
	structure_new1.replace(stacks[i][0], stacks[i][1] - stacks[i][0] + 1, string(stacks[i][1] - stacks[i][0] + 1, '.'));
	append_stack(stacks[i], structure_new1);
	string constraint1 = unconstrained_matrix[i].str_1;
	

	string structure_new2(sequence.length(), 'x');
	structure_new2.replace(stacks[j][0], stacks[j][1] - stacks[j][0] + 1, string(stacks[j][1] - stacks[j][0] + 1, '.'));
	append_stack(stacks[j], structure_new2);
	string constraint2 = unconstrained_matrix[j].str_1;
	
	int energy_T1_1 = stacks[i][3] + SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i])
		+ energy_model.FE_hairpin(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1)
		+ SHAPE_constraint.evaluate_pseudo_energy(stacks[i][0], stacks[i][1], structure_new1, constraint2, frequency);
	int energy_T1_2 = unconstrained_matrix[j].free_energy_1;
	
	int energy_T2_1 = unconstrained_matrix[i].free_energy_1;
	int energy_T2_2 = stacks[j][3] + SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[j]) 
		+ energy_model.FE_hairpin(stacks[j][0] + stacks[j][2] - 1, stacks[j][1] - stacks[j][2] + 1)
		+ SHAPE_constraint.evaluate_pseudo_energy(stacks[j][0], stacks[j][1], structure_new2, constraint1, 100 - frequency);
	
	if(energy_T1_1 + energy_T1_2 < energy_T2_1 + energy_T2_2)	{
		record_hairpin.free_energy_1 = energy_T1_1;
		record_hairpin.free_energy_2 = energy_T1_2;
		record_hairpin.sum_free_energy = record_hairpin.free_energy_1 + record_hairpin.free_energy_2;
		compress_structure(0, structure_new1, record_hairpin);
		compress_structure(1, constraint2, record_hairpin);
	}	else	{
		record_hairpin.free_energy_1 = energy_T2_1;
		record_hairpin.free_energy_2 = energy_T2_2;
		record_hairpin.sum_free_energy = record_hairpin.free_energy_1 + record_hairpin.free_energy_2;
		compress_structure(0, constraint1, record_hairpin);
		compress_structure(1, structure_new2, record_hairpin);
	}
	
	//cout << structure_new1 << endl;
	//cout << structure_new2 << endl;
	//cout << "************" << endl;
	
	return;
}

void bi_fold_stack::fold_internal(int i, int j, folding_matrix &record_internal)	{
	//cout << i << "	" << j << endl;
	//cout << constrained_matrix[i][j].str_begin_1 << "	" << constrained_matrix[i][j].str_end_1 << endl;
	//cout << constrained_matrix[i][j].str_begin_2 << "	" << constrained_matrix[i][j].str_end_2 << endl;
	
	int m, n;
	energy_parameters energy_model(sequence);
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	string constraint1, constraint2;
	record_internal.sum_free_energy = LARGENUM;
	
	for(m = num_internal[i] - 1; m >= 0; -- m)	{
		int x = internal[i][m];
		/* construct the new structure */
		string structure_new(sequence.length(), 'x');
		int len = stacks[i][1] - stacks[i][0] - 2 * stacks[i][2] + 1;
		structure_new.replace(stacks[i][0] + stacks[i][2], len, string(len, '.'));
		append_stack(stacks[i], structure_new);
		structure_new.replace(stacks[x][0], stacks[x][1] - stacks[x][0] + 1, 
			constrained_matrix[x][j].str_1.substr(stacks[x][0], stacks[x][1] - stacks[x][0] + 1));
		/* retrieve the constraint */
		constraint2 = constrained_matrix[x][j].str_2;
		/* computing the new energy */
		int energy_new =  constrained_matrix[x][j].free_energy_1 + stacks[i][3]
			+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i]) 
			+ energy_model.FE_loop(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1, stacks[x][1], stacks[x][0])
			+ SHAPE_constraint.evaluate_pseudo_energy_exclude(stacks[i][0], stacks[i][1], stacks[x][0], stacks[x][1], structure_new, constraint2, frequency);
		//cout << stacks[i][0] + stacks[i][2] - 1 << "	" <<  stacks[i][1] - stacks[i][2] + 1 << "	" << stacks[x][1] << "	" << stacks[x][0] << endl;
		//cout << record_internal.free_energy_1 << endl;
		//cout << "******************" << endl;
		if(energy_new + constrained_matrix[x][j].free_energy_2 < record_internal.sum_free_energy)	{
			copy_folding_matrix(constrained_matrix[x][j], record_internal);
			record_internal.free_energy_1 = energy_new;
			record_internal.sum_free_energy = record_internal.free_energy_1 + record_internal.free_energy_2;
			compress_structure(0, structure_new, record_internal);
		}
	}
	//cout << record_internal.sum_free_energy << endl;
	for(n = num_internal[j] - 1; n >= 0; -- n)	{
		int y = internal[j][n];
		/* construct the new structure */
		string structure_new(sequence.length(), 'x');
		int len = stacks[j][1] - stacks[j][0] - 2 * stacks[j][2] + 1;
		structure_new.replace(stacks[j][0] + stacks[j][2], len, string(len, '.'));
		append_stack(stacks[j], structure_new);	
		structure_new.replace(stacks[y][0], stacks[y][1] - stacks[y][0] + 1, 
			constrained_matrix[i][y].str_2.substr(stacks[y][0], stacks[y][1] - stacks[y][0] + 1));
		/* retrieve the constraint */
		constraint1 = constrained_matrix[i][y].str_1;
		/* computing the new energy */
		int energy_new = constrained_matrix[i][y].free_energy_2 + stacks[j][3]
			+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[j]) 
			+ energy_model.FE_loop(stacks[j][0] + stacks[j][2] - 1, stacks[j][1] - stacks[j][2] + 1, stacks[y][1], stacks[y][0])
			+ SHAPE_constraint.evaluate_pseudo_energy_exclude(stacks[j][0], stacks[j][1], stacks[y][0], stacks[y][1], structure_new, constraint1, 100 - frequency);
		//cout << stacks[i][0] + stacks[i][2] - 1 << "	" <<  stacks[i][1] - stacks[i][2] + 1 << "	" << stacks[x][1] << "	" << stacks[x][0] << endl;
		//cout << record_internal.free_energy_1 << endl;
		//cout << "******************" << endl;
		if(energy_new + constrained_matrix[i][y].free_energy_1 < record_internal.sum_free_energy)	{
			copy_folding_matrix(constrained_matrix[i][y], record_internal);
			record_internal.free_energy_2 = energy_new;
			record_internal.sum_free_energy = record_internal.free_energy_1 + record_internal.free_energy_2;
			compress_structure(1, structure_new, record_internal);
		}
	}
	//cout << record_internal.sum_free_energy << endl;
	//cout << "**************" << endl;
	return;
}

void bi_fold_stack::optimal_configuration(bool which, int i, folding_matrix constraint, folding_matrix &optimal)	{
	int m, n;
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	if(which == 0)	{
		optimal.free_energy_1 = unconstrained_matrix[i].free_energy_1;
		optimal.str_begin_1 = unconstrained_matrix[i].str_begin_1;
		optimal.str_end_1 = unconstrained_matrix[i].str_end_1;
		optimal.str_stacks_1 = unconstrained_matrix[i].str_stacks_1; 
		optimal.str_1 = unconstrained_matrix[i].str_1; 
		/* find the constraint structure */
		string str_constraint = constraint.str_2;
		for(m = 0; m < constraint.str_stacks_2.size(); ++ m)	{
			int x = constraint.str_stacks_2[m];
			int min_partial_energy = LARGENUM;
			/* check whether they overlap */
			if(is_overlap(stacks[i], stacks[x]))	{
				/* evaluate pseudo energy for non-overlapped regions */
				int real_energy = constrained_matrix[i][x].real_energy_1;	
				int optimistic_energy = SHAPE_constraint.optimistic_pseudo_energy_structure(constrained_matrix[i][x].str_stacks_1, stacks);
				/*
				if(which == 0 && i == 10 && x == 7)	{
					cout << "i:	" << i << "	x:	" << x << endl;
					cout << "energy:	" << real_energy << endl;
					cout << "optimistic energy:	" << optimistic_energy << endl;
					cout << decompress_structure(0, constrained_matrix[i][x]) << endl;
					cout << decompress_structure(1, constrained_matrix[i][x]) << endl;
				}
				*/	
				if(real_energy + optimistic_energy < min_partial_energy)	{
					min_partial_energy = real_energy + optimistic_energy;
					optimal.free_energy_1 = min_partial_energy;
					optimal.str_begin_1 = constrained_matrix[i][x].str_begin_1;
					optimal.str_end_1 = constrained_matrix[i][x].str_end_1;
					optimal.str_stacks_1 = constrained_matrix[i][x].str_stacks_1;
					optimal.str_1 = constrained_matrix[i][x].str_1;
				}
			}
		}
	}	else	{
		optimal.free_energy_2 = unconstrained_matrix[i].free_energy_1;
		optimal.str_begin_2 = unconstrained_matrix[i].str_begin_1;
		optimal.str_end_2 = unconstrained_matrix[i].str_end_1;
		optimal.str_stacks_2 = unconstrained_matrix[i].str_stacks_1; 
		optimal.str_2 = unconstrained_matrix[i].str_1; 
		/* find the constraint structure */
		string str_constraint = constraint.str_1;
		for(m = 0; m < constraint.str_stacks_1.size(); ++ m)	{
			int x = constraint.str_stacks_1[m];
			int min_partial_energy = LARGENUM;
			/* check whether they overlap */
			if(is_overlap(stacks[x], stacks[i]))	{
				/* evaluate pseudo energy for non-overlapped regions */
				int real_energy = constrained_matrix[x][i].real_energy_2;
				int optimistic_energy = SHAPE_constraint.optimistic_pseudo_energy_structure(constrained_matrix[x][i].str_stacks_2, stacks);
					
				if(real_energy + optimistic_energy < min_partial_energy)	{
					min_partial_energy = real_energy + optimistic_energy;
					optimal.free_energy_2 = min_partial_energy;
					optimal.str_begin_2 = constrained_matrix[x][i].str_begin_2;
					optimal.str_end_2 = constrained_matrix[x][i].str_end_2;
					optimal.str_stacks_2 = constrained_matrix[x][i].str_stacks_2;
					optimal.str_2 = constrained_matrix[x][i].str_2;
				}
			}
		}
	}
	return;
}

void bi_fold_stack::fill_chain_matrix(int i, int j, vector< vector<folding_matrix> > &chain_matrix)	{
	
	int k, l, m, n;
	energy_parameters energy_model_i(sequence);
	if(i == stacks.size() - 1)	{
		energy_model_i.set_pseudo_stack_energy();
	}
	energy_parameters energy_model_j(sequence);
	if(j == stacks.size() - 1)	{
		energy_model_j.set_pseudo_stack_energy();
	}
	
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	/*
	vector < vector<folding_matrix> > chain_matrix (stacks.size());
	for(m = 0; m < stacks.size(); ++ m)	{
		chain_matrix[m] = vector<folding_matrix> (stacks.size());
	}
	*/
	//cout << "begin of fold_multi:	" << i << "	" << j << endl;
	
	/* folding the optimal folding between two chains */
	int x, y;
	for(m = num_enclosing[i] - 1; m >= 0; -- m)	{
		x = enclosing[i][m];
		for(n = num_enclosing[j] - 1; n >= 0; -- n)	{
			y = enclosing[j][n];
			/* case 1: x constrained by E[Fset[i], y] */
			folding_matrix x_constrained;
			x_constrained.sum_free_energy = LARGENUM;
			for(k = num_Fset[x] - 1; k >= 0; -- k)	{
				int xprev = Fset[x][k];
				if(stacks[xprev][0] >= stacks[i][0] + stacks[i][2])	{
					//cout << "finding optimal configuration" << endl;
					folding_matrix traverse_Fset;
					optimal_configuration(0, x, chain_matrix[xprev][y], traverse_Fset);
					//cout << "********************" << endl;
					//cout << x << "	" << y << "	" << xprev << "	" << y << endl;
					//cout << decompress_structure(0, traverse_Fset) << endl;
					//cout << "********************" << endl;
					string str_x = traverse_Fset.str_1;
					string str_y = chain_matrix[xprev][y].str_2;
					int constraint_energy = SHAPE_constraint.evaluate_pseudo_energy(stacks[xprev][1] + 1, stacks[x][1], str_x, str_y, frequency); 
					if(traverse_Fset.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy < 0 && 
						traverse_Fset.free_energy_1 + chain_matrix[xprev][y].sum_free_energy + energy_model_i.Multi_new_branch + constraint_energy
						< x_constrained.sum_free_energy)	{
						/* the stack x is taken */
						x_constrained.free_energy_1 = chain_matrix[xprev][y].free_energy_1 
							+ traverse_Fset.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy;
						x_constrained.free_energy_2 = chain_matrix[xprev][y].free_energy_2;
						x_constrained.sum_free_energy = x_constrained.free_energy_1 + x_constrained.free_energy_2;
						/* set structure for the first transcript */
						string str_new = decompress_structure(0, chain_matrix[xprev][y]);
						str_new.replace(stacks[xprev][1] + 1, stacks[x][0] - stacks[xprev][1] - 1, 
							string(stacks[x][0] - stacks[xprev][1] - 1, '.'));
						str_new.replace(stacks[x][0], stacks[x][1] - stacks[x][0] + 1, 
							traverse_Fset.str_1.substr(stacks[x][0], stacks[x][1] - stacks[x][0] + 1));
						compress_structure(0, str_new, x_constrained);
						/* set structure for the second transcript */
						x_constrained.str_begin_2 = chain_matrix[xprev][y].str_begin_2;
						x_constrained.str_end_2 = chain_matrix[xprev][y].str_end_2;
						x_constrained.str_stacks_2 = chain_matrix[xprev][y].str_stacks_2;
					}	else if(traverse_Fset.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy >= 0 &&
						chain_matrix[xprev][y].sum_free_energy < x_constrained.sum_free_energy)	{
						/* the stack x is not taken */
						x_constrained.free_energy_1 = chain_matrix[xprev][y].free_energy_1;
						x_constrained.free_energy_2 = chain_matrix[xprev][y].free_energy_2;
						x_constrained.sum_free_energy = x_constrained.free_energy_1 + x_constrained.free_energy_2;
						/* set structure for the first transcript */
						x_constrained.str_begin_1 = chain_matrix[xprev][y].str_begin_1;
						x_constrained.str_end_1 = stacks[x][1];
						x_constrained.str_stacks_1 = chain_matrix[xprev][y].str_stacks_1;
						/* set structure for the second transcript */
						x_constrained.str_begin_2 = chain_matrix[xprev][y].str_begin_2;
						x_constrained.str_end_2 = chain_matrix[xprev][y].str_end_2;
						x_constrained.str_stacks_2 = chain_matrix[xprev][y].str_stacks_2;
					}
					
					/*
					if(i == 74 && j == 80 && x == 24 && y == 18)	{
						cout << "###################" << endl;
						cout << "xprev:	" << xprev << endl;
						cout << "chain_matrix energy 1:	" << chain_matrix[xprev][y].free_energy_1 << endl;
						cout << "chain_matrix energy 2:	" << chain_matrix[xprev][y].free_energy_2 << endl;
						cout << "traverse energy_2:	" << traverse_Fset.free_energy_1 << endl;
						cout << "loop energy:	" << energy_model_i.Multi_new_branch << endl;
						cout << "constraint energy:	" << SHAPE_constraint.evaluate_pseudo_energy(stacks[x][0], stacks[x][1], str_x, str_y, frequency) << endl;
						cout << "x_constrianed energy:	" << x_constrained.sum_free_energy << endl;					
					}
					*/
				}
			}
			//cout << "finished x_constrained" << endl;
			
			/* case 2: y constrained by E[x, Fset[j]] */
			folding_matrix y_constrained;
			y_constrained.sum_free_energy = LARGENUM;
			for(k = num_Fset[y] - 1; k >= 0; -- k)	{
				int yprev = Fset[y][k];
				if(stacks[yprev][0] >= stacks[j][0] + stacks[j][2])	{
					
					folding_matrix traverse_Fset;
					optimal_configuration(1, y, chain_matrix[x][yprev], traverse_Fset);
					string str_x = chain_matrix[x][yprev].str_1;
					string str_y = traverse_Fset.str_2;
					int constraint_energy = SHAPE_constraint.evaluate_pseudo_energy(stacks[yprev][1], stacks[y][1], str_y, str_x, 100 - frequency);
					if(traverse_Fset.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy < 0 && 
						(traverse_Fset.free_energy_2 + chain_matrix[x][yprev].sum_free_energy 
							+ energy_model_j.Multi_new_branch + constraint_energy < y_constrained.sum_free_energy))	{
						
						/* the stack x is taken */
						y_constrained.free_energy_1 = chain_matrix[x][yprev].free_energy_1;
						y_constrained.free_energy_2 = chain_matrix[x][yprev].free_energy_2 
							+ traverse_Fset.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy;
						y_constrained.sum_free_energy = y_constrained.free_energy_1 + y_constrained.free_energy_2;
						/* set structure for the first transcript */
						y_constrained.str_begin_1 = chain_matrix[x][yprev].str_begin_1;
						y_constrained.str_end_1 = chain_matrix[x][yprev].str_end_1;
						y_constrained.str_stacks_1 = chain_matrix[x][yprev].str_stacks_1;
						/* set structure for the second transcript */
						string str_new = chain_matrix[x][yprev].str_2;
						str_new.replace(stacks[yprev][1] + 1, stacks[y][0] - stacks[yprev][1] - 1, 
							string(stacks[y][0] - stacks[yprev][1] - 1, '.'));
						str_new.replace(stacks[y][0], stacks[y][1] - stacks[y][0] + 1, 
							traverse_Fset.str_2.substr(stacks[y][0], stacks[y][1] - stacks[y][0] + 1));
						compress_structure(1, str_new, y_constrained);
					}	else if(traverse_Fset.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy >= 0 &&
						chain_matrix[x][yprev].sum_free_energy < y_constrained.sum_free_energy)	{
						/* the stack x is not taken */
						y_constrained.free_energy_1 = chain_matrix[x][yprev].free_energy_1;
						y_constrained.free_energy_2 = chain_matrix[x][yprev].free_energy_2;
						y_constrained.sum_free_energy = y_constrained.free_energy_1 + y_constrained.free_energy_2;
						/* set structure for the first transcript */
						y_constrained.str_begin_1 = chain_matrix[x][yprev].str_begin_1;
						y_constrained.str_end_1 = chain_matrix[x][yprev].str_end_1;
						y_constrained.str_stacks_1 = chain_matrix[x][yprev].str_stacks_1;
						/* set structure for the second transcript */
						y_constrained.str_begin_2 = chain_matrix[x][yprev].str_begin_2;
						y_constrained.str_end_2 = stacks[y][1];
						y_constrained.str_stacks_2 = chain_matrix[x][yprev].str_stacks_2;
					}
					/*
					if(i == 74 && j == 80 && x == 18 && y == 18)	{
						cout << "()()()()()()()()" << endl;
						cout << "yprev:	" << yprev << endl;
						cout << "chain_matrix energy 1:	" << chain_matrix[x][yprev].free_energy_1 << endl;
						cout << "chain_matrix energy 2:	" << chain_matrix[x][yprev].free_energy_2 << endl;
						cout << "traverse energy_2:	" << traverse_Fset.free_energy_2 << endl;
						cout << "loop energy:	" << energy_model_j.Multi_new_branch << endl;
						cout << "constraint energy:	" << SHAPE_constraint.evaluate_pseudo_energy(stacks[y][0], stacks[y][1], str_y, str_x, 100 - frequency) << endl;
						cout << "y_constrianed energy:	" << y_constrained.sum_free_energy << endl;					
					}
					*/
				}
			}
			
			//cout << "finished y_constrained" << endl;
			//cout << "**********************" << endl;
			//cout << i << "	" << j << "	" << x << "	" << y << endl;
			//cout << "begin of newly wirtten" << endl;
			
			/* case 3: a boundary case: where x is the first stack */
			folding_matrix x_boundary;
			folding_matrix traverse_boundary_x;
			optimal_configuration(0, x, unconstrained_multi_matrix[j][y], traverse_boundary_x);
			string str_x = traverse_boundary_x.str_1;
			string str_y = unconstrained_multi_matrix[j][y].str_2;
			int constraint_energy = SHAPE_constraint.evaluate_pseudo_energy(stacks[i][0] + stacks[i][2], stacks[x][1], str_x, str_y, frequency);
			if(traverse_boundary_x.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy < 0)	{
				x_boundary.free_energy_1 = traverse_boundary_x.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy;
				x_boundary.free_energy_2 = unconstrained_multi_matrix[j][y].free_energy_1;
				x_boundary.sum_free_energy = x_boundary.free_energy_1 + x_boundary.free_energy_2;
				x_boundary.str_begin_1 = stacks[i][0] + stacks[i][2];
				x_boundary.str_end_1 = stacks[x][1];
				x_boundary.str_stacks_1 = traverse_boundary_x.str_stacks_1;
				x_boundary.str_begin_2 = unconstrained_multi_matrix[j][y].str_begin_1;
				x_boundary.str_end_2 = unconstrained_multi_matrix[j][y].str_end_1;
				x_boundary.str_stacks_2 = unconstrained_multi_matrix[j][y].str_stacks_1;
			}	else	{
				x_boundary.free_energy_1 = 0;
				x_boundary.free_energy_2 = unconstrained_multi_matrix[j][y].free_energy_1;
				x_boundary.sum_free_energy = x_boundary.free_energy_1 + x_boundary.free_energy_2;
				x_boundary.str_begin_1 = stacks[i][0] + stacks[i][2];
				x_boundary.str_end_1 = stacks[x][1];
				x_boundary.str_begin_2 = unconstrained_multi_matrix[j][y].str_begin_1;
				x_boundary.str_end_2 = unconstrained_multi_matrix[j][y].str_end_1;
				x_boundary.str_stacks_2 = unconstrained_multi_matrix[j][y].str_stacks_1;
			}
			
			
			/* case 4: a boundary case: where y is the first stack */
			folding_matrix y_boundary;
			folding_matrix traverse_boundary_y;
			optimal_configuration(1, y, unconstrained_multi_matrix[i][x], traverse_boundary_y);
			str_x = unconstrained_multi_matrix[i][x].str_1;
			str_y = traverse_boundary_y.str_2;
			constraint_energy = SHAPE_constraint.evaluate_pseudo_energy(stacks[j][0] + stacks[j][2], stacks[y][1], str_y, str_x, 100 - frequency);
			if(traverse_boundary_y.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy < 0)	{
				y_boundary.free_energy_1 = unconstrained_multi_matrix[i][x].free_energy_1;
				y_boundary.free_energy_2 = traverse_boundary_y.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy;
				y_boundary.sum_free_energy = y_boundary.free_energy_1 + y_boundary.free_energy_2;
				y_boundary.str_begin_1 = unconstrained_multi_matrix[i][x].str_begin_1;
				y_boundary.str_end_1 = unconstrained_multi_matrix[i][x].str_end_1;
				y_boundary.str_stacks_1 = unconstrained_multi_matrix[i][x].str_stacks_1;
				y_boundary.str_begin_2 = stacks[j][0] + stacks[j][2];
				y_boundary.str_end_2 = stacks[y][1];
				y_boundary.str_stacks_2 = traverse_boundary_y.str_stacks_2;
			}	else	{
				y_boundary.free_energy_1 = unconstrained_multi_matrix[i][x].free_energy_1;
				y_boundary.free_energy_2 = 0;
				y_boundary.sum_free_energy = y_boundary.free_energy_1 + y_boundary.free_energy_2;
				y_boundary.str_begin_1 = unconstrained_multi_matrix[i][x].str_begin_1;
				y_boundary.str_end_1 = unconstrained_multi_matrix[i][x].str_end_1;
				y_boundary.str_stacks_1 = unconstrained_multi_matrix[i][x].str_stacks_1;
				y_boundary.str_begin_2 = stacks[j][0] + stacks[j][2];
				y_boundary.str_end_2 = stacks[y][1];
			}
			
			/* case for both non structured */
			folding_matrix both_unstructured;
			both_unstructured.free_energy_1 = both_unstructured.free_energy_2 = both_unstructured.sum_free_energy = 0;
			both_unstructured.str_begin_1 = stacks[i][0] + stacks[i][2];
			both_unstructured.str_end_1 = stacks[x][1];
			both_unstructured.str_begin_2 = stacks[j][0] + stacks[j][2];
			both_unstructured.str_end_2 = stacks[y][1];
			
			
			//cout << "finished special cases" << endl;
			
			if(x_constrained.sum_free_energy <= y_constrained.sum_free_energy && 
				x_constrained.sum_free_energy <= x_boundary.sum_free_energy && 
				x_constrained.sum_free_energy <= y_boundary.sum_free_energy && 
				x_constrained.sum_free_energy <= 0)	{
				/*
				if(i == 74 && j == 80 && x == 24 && y == 18)	{
					cout << "chain_matrix:	x_constrained" << endl;
				}
				*/
				copy_folding_matrix(x_constrained, chain_matrix[x][y]);
			
			}	else if(y_constrained.sum_free_energy <= x_constrained.sum_free_energy && 
				y_constrained.sum_free_energy <= x_boundary.sum_free_energy && 
				y_constrained.sum_free_energy <= y_boundary.sum_free_energy && 
				y_constrained.sum_free_energy <= 0)	{
				
				/*
				if(i == 74 && j == 80 && x == 24 && y == 18)	{
					cout << "chain_matrix:	y_constrained" << endl;
				}
				*/
				copy_folding_matrix(y_constrained, chain_matrix[x][y]);
			
			}	else if(x_boundary.sum_free_energy <= x_constrained.sum_free_energy && 
				x_boundary.sum_free_energy <= y_constrained.sum_free_energy && 
				x_boundary.sum_free_energy <= y_boundary.sum_free_energy && 
				x_boundary.sum_free_energy <= 0)	{
				
				/*
				if(i == 74 && j == 80 && x == 24 && y == 18)	{
					cout << "chain_matrix:	x_boundary" << endl;
				}
				*/
				copy_folding_matrix(x_boundary, chain_matrix[x][y]);
			
			}	else if(y_boundary.sum_free_energy <= x_constrained.sum_free_energy && 
				y_boundary.sum_free_energy <= y_constrained.sum_free_energy && 
				y_boundary.sum_free_energy <= x_boundary.sum_free_energy && 
				y_boundary.sum_free_energy <= 0)	{
				
				/*
				if(i == 74 && j == 80 && x == 24 && y == 18)	{
					cout << "chain_matrix:	y_boundary" << endl;
				}
				*/
				copy_folding_matrix(y_boundary, chain_matrix[x][y]);
			
			}	else	{
				
				/*
				if(i == 74 && j == 80 && x == 24 && y == 18)	{
					cout << "chain_matrix:	both unstructured" << endl;
				}
				*/
				copy_folding_matrix(both_unstructured, chain_matrix[x][y]);
			
			}
			
			chain_matrix[x][y].str_1 = decompress_structure(0, chain_matrix[x][y]);
			chain_matrix[x][y].str_2 = decompress_structure(1, chain_matrix[x][y]);
			/*
			if(i == 38 && j == 57 && x == 23 && y == 43)	{
				cout << "$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$" << endl;
				cout << "x and y:	" << x << "	" << y << endl;
				cout << "energy_1:	" << chain_matrix[x][y].free_energy_1 << endl;
				cout << "evaluate:	" << evaluate_energy_stack(chain_matrix[x][y].str_stacks_1) << endl;
				cout << decompress_structure(0, chain_matrix[x][y]) << endl;
				
				cout << "energy_2:	" << chain_matrix[x][y].free_energy_2 << endl;
				cout << "evaluate:	" << evaluate_energy_stack(chain_matrix[x][y].str_stacks_2) << endl;
				cout << decompress_structure(1, chain_matrix[x][y]) << endl;
				cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
			}
			*/
			//cout << "**************" << endl;
			//cout << decompress_structure(0, chain_matrix[x][y]) << endl;
			//cout << decompress_structure(1, chain_matrix[x][y]) << endl;
			/*
			if(i == 80 && j == 80 && x == 68 && y == 53)	{
				cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
				cout << "x and y:	" << x << "	" << y << endl;
				cout << "energy_1:	" << chain_matrix[x][y].free_energy_1 << endl;
				cout << "evaluate:	" << evaluate_energy_stack(chain_matrix[x][y].str_stacks_1) << endl;
				cout << decompress_structure(0, chain_matrix[x][y]) << endl;
				
				cout << "energy_2:	" << chain_matrix[x][y].free_energy_2 << endl;
				cout << "evaluate:	" << evaluate_energy_stack(chain_matrix[x][y].str_stacks_2) << endl;
				cout << decompress_structure(1, chain_matrix[x][y]) << endl;
				cout << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
			}
			*/
		}
	}

	return;
}

void bi_fold_stack::fill_chainloop_matrix_iloop(int i, int j, vector< vector<folding_matrix> > chain_matrix)	{
	
	int k, l, m, n;
	energy_parameters energy_model_i(sequence);
	energy_parameters energy_model_j(sequence);
	if(i == stacks.size() - 1)	{
		energy_model_i.set_pseudo_stack_energy();
	}
	if(j == stacks.size() - 1)	{
		energy_model_j.set_pseudo_stack_energy();
	}
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	
	for(l = num_enclosing[j] - 1; l >= 0; -- l)	{
		int y = enclosing[j][l];
		//if(i == 0 && j == 2)	{
		//	cout << "y:	" << y << endl;
		//}
		
		/* y constrained */
		folding_matrix y_constrained;
		y_constrained.sum_free_energy = LARGENUM;
		for(n = num_Fset[y] - 1; n >= 0; -- n)	{
			int yprev = Fset[y][n];
			if(stacks[yprev][0] >= stacks[j][0] + stacks[j][2])	{
				folding_matrix traverse_Fset;
				optimal_configuration(1, y, chainloop_matrix_iloop[i][j][yprev], traverse_Fset);
				string str_x = chainloop_matrix_iloop[i][j][yprev].str_1;
				string str_y = traverse_Fset.str_2;
				int constraint_energy = SHAPE_constraint.evaluate_pseudo_energy(stacks[yprev][1] + 1, stacks[y][1], str_y, str_x, 100 - frequency);
				if(traverse_Fset.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy < 0 && 
					chainloop_matrix_iloop[i][j][yprev].sum_free_energy + traverse_Fset.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy 
					< y_constrained.sum_free_energy)	{
					/* y taking the configuration as indicated by traverse_Fset */
					y_constrained.free_energy_1 = chainloop_matrix_iloop[i][j][yprev].free_energy_1;
					y_constrained.free_energy_2 = chainloop_matrix_iloop[i][j][yprev].free_energy_2
						+ traverse_Fset.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy;
					y_constrained.sum_free_energy = y_constrained.free_energy_1 + y_constrained.free_energy_2;
					/* fillup structures */
					y_constrained.str_begin_1 = chainloop_matrix_iloop[i][j][yprev].str_begin_1;
					y_constrained.str_end_1 = chainloop_matrix_iloop[i][j][yprev].str_end_1;
					y_constrained.str_stacks_1 = chainloop_matrix_iloop[i][j][yprev].str_stacks_1;
					string str_new = chainloop_matrix_iloop[i][j][yprev].str_2;
					str_new.replace(stacks[yprev][1] + 1, stacks[y][0] - stacks[yprev][1] - 1,
						string(stacks[y][0] - stacks[yprev][1] - 1, '.'));
					str_new.replace(stacks[y][0], stacks[y][1] - stacks[y][0] + 1, 
						traverse_Fset.str_2.substr(stacks[y][0], stacks[y][1] - stacks[y][0] + 1));
					compress_structure(1, str_new, y_constrained);
				}	else if(traverse_Fset.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy >= 0 && 
					chainloop_matrix_iloop[i][j][yprev].sum_free_energy < y_constrained.sum_free_energy)	{
					/* y taking no structure */
					y_constrained.free_energy_1 = chainloop_matrix_iloop[i][j][yprev].free_energy_1;
					y_constrained.free_energy_2 = chainloop_matrix_iloop[i][j][yprev].free_energy_2;
					y_constrained.sum_free_energy = chainloop_matrix_iloop[i][j][yprev].sum_free_energy;
					/* fillup structures */
					y_constrained.str_begin_1 = chainloop_matrix_iloop[i][j][yprev].str_begin_1;
					y_constrained.str_end_1 = chainloop_matrix_iloop[i][j][yprev].str_end_1;
					y_constrained.str_stacks_1 = chainloop_matrix_iloop[i][j][yprev].str_stacks_1;
					y_constrained.str_begin_2 = chainloop_matrix_iloop[i][j][yprev].str_begin_2;
					y_constrained.str_end_2 = stacks[y][1];
					y_constrained.str_stacks_2 = chainloop_matrix_iloop[i][j][yprev].str_stacks_2;
				}
			}
		}
		//if(i == 0 && j == 2)	{
		//	cout << "y_constrained:	" << y_constrained.sum_free_energy << endl;
		//}
		
		/* y boundary */
		folding_matrix y_boundary;
		folding_matrix traverse_boundary_y;
		optimal_configuration(1, y, unconstrained_matrix[i], traverse_boundary_y);
		string str_x = unconstrained_matrix[i].str_1;
		string str_y = traverse_boundary_y.str_2;
		int constraint_energy = SHAPE_constraint.evaluate_pseudo_energy(stacks[j][0] + stacks[j][2], stacks[y][1], str_y, str_x, 100 - frequency);
		if(traverse_boundary_y.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy < 0)	{
			y_boundary.free_energy_1 = unconstrained_matrix[i].free_energy_1;
			y_boundary.free_energy_2 = traverse_boundary_y.free_energy_2 + energy_model_j.Multi_new_branch + constraint_energy;
			y_boundary.sum_free_energy = y_boundary.free_energy_1 + y_boundary.free_energy_2;
			y_boundary.str_begin_1 = unconstrained_matrix[i].str_begin_1;
			y_boundary.str_end_1 = unconstrained_matrix[i].str_end_1;
			y_boundary.str_stacks_1 = unconstrained_matrix[i].str_stacks_1;
			y_boundary.str_begin_2 = stacks[j][0] + stacks[j][2];
			y_boundary.str_end_2 = stacks[y][1];
			y_boundary.str_stacks_2 = traverse_boundary_y.str_stacks_2;
		}	else	{
			y_boundary.free_energy_1 = unconstrained_matrix[i].free_energy_1;
			y_boundary.free_energy_2 = 0;
			y_boundary.sum_free_energy = y_boundary.free_energy_1 + y_boundary.free_energy_2;
			y_boundary.str_begin_1 = unconstrained_matrix[i].str_begin_1;
			y_boundary.str_end_1 = unconstrained_matrix[i].str_end_1;
			y_boundary.str_stacks_1 = unconstrained_matrix[i].str_stacks_1;
			y_boundary.str_begin_2 = stacks[j][0] + stacks[j][2];
			y_boundary.str_end_2 = stacks[y][1];
		}
		
		//if(i == 0 && j == 2)	{
		//	cout << "y_boundary:	" << y_boundary.sum_free_energy << endl;
		//}
		
			
		/* i as hairpin loop */
		folding_matrix i_hairpin;
		string str_hairpin_i(sequence.length(), 'x');
		str_hairpin_i.replace(stacks[i][0], stacks[i][1] - stacks[i][0] + 1, string(stacks[i][1] - stacks[i][0] + 1, '.'));
		append_stack(stacks[i], str_hairpin_i);
		string str_chain_y = unconstrained_multi_matrix[j][y].str_1;
		i_hairpin.free_energy_1 = stacks[i][3] + energy_model_i.FE_hairpin(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1)
			+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i])
			+ SHAPE_constraint.evaluate_pseudo_energy(stacks[i][0], stacks[i][1], str_hairpin_i, str_chain_y, frequency);
		i_hairpin.free_energy_2 = unconstrained_multi_matrix[j][y].free_energy_2;
		i_hairpin.sum_free_energy = i_hairpin.free_energy_1 + i_hairpin.free_energy_2;
		i_hairpin.str_begin_1 = stacks[i][0];
		i_hairpin.str_end_1 = stacks[i][1];
		i_hairpin.str_stacks_1.resize(1);
		i_hairpin.str_stacks_1[0] = i;
		i_hairpin.str_begin_2 = unconstrained_multi_matrix[j][y].str_begin_1;
		i_hairpin.str_end_2 = unconstrained_multi_matrix[j][y].str_end_1;
		i_hairpin.str_stacks_2 = unconstrained_multi_matrix[j][y].str_stacks_1;
		
		//if(i == 0 && j == 2)	{
		//	cout << "i_hairpin:	" << endl;
		//}
		
		
		/* i as internal loop */
		folding_matrix i_internal;
		i_internal.sum_free_energy = LARGENUM;
		for(m = num_internal[i] - 1; m >= 0; -- m)	{
			int ien = internal[i][m];
			string str_internal_i(sequence.length(), 'x');
			str_internal_i.replace(stacks[i][0], stacks[i][1] - stacks[i][0] + 1, string(stacks[i][1] - stacks[i][0] + 1, '.'));
			append_stack(stacks[i], str_internal_i);
			str_internal_i.replace(stacks[ien][0], stacks[ien][1] - stacks[ien][0] + 1, 
				chainloop_matrix_iloop[ien][j][y].str_1.substr(stacks[ien][0], stacks[ien][1] - stacks[ien][0] + 1));
			str_chain_y = chainloop_matrix_iloop[ien][j][y].str_2;
			int energy_new = chainloop_matrix_iloop[ien][j][y].sum_free_energy + stacks[i][3]
				+ energy_model_i.FE_loop(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1, stacks[ien][1], stacks[ien][0])
				+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i])
				+ SHAPE_constraint.evaluate_pseudo_energy_exclude(stacks[i][0], stacks[i][1], stacks[ien][0], stacks[ien][1], str_internal_i, str_chain_y, frequency);
			if(energy_new < i_internal.sum_free_energy)	{
				i_internal.free_energy_1 = energy_new - chainloop_matrix_iloop[ien][j][y].free_energy_2;
				i_internal.free_energy_2 = chainloop_matrix_iloop[ien][j][y].free_energy_2;
				i_internal.sum_free_energy = energy_new;
				i_internal.str_begin_1 = stacks[i][0];
				i_internal.str_end_1 = stacks[i][1];
				i_internal.str_stacks_1 = chainloop_matrix_iloop[ien][j][y].str_stacks_1;
				i_internal.str_stacks_1.resize(i_internal.str_stacks_1.size() + 1);
				i_internal.str_stacks_1[i_internal.str_stacks_1.size() - 1] = i;
				i_internal.str_begin_2 = chainloop_matrix_iloop[ien][j][y].str_begin_2;
				i_internal.str_end_2 = chainloop_matrix_iloop[ien][j][y].str_end_2;
				i_internal.str_stacks_2 = chainloop_matrix_iloop[ien][j][y].str_stacks_2;
			}
		}
		//if(i == 0 && j == 2)	{
		//	cout << "i_internal:	" << endl;
		//}
		
		
		/* i as multi loop */
		folding_matrix i_multi;
		i_multi.sum_free_energy = LARGENUM;
		for(m = num_FPset[i] - 1; m >= 0; -- m)	{
			int x = FPset[i][m];
			string str_multi_i(sequence.length(), 'x');
			str_multi_i.replace(stacks[i][0], stacks[i][1] - stacks[i][0] + 1, string(stacks[i][1] - stacks[i][0] + 1, '.'));
			append_stack(stacks[i], str_multi_i);
			str_multi_i.replace(stacks[i][0] + stacks[i][2], stacks[x][1] - stacks[i][0] - stacks[i][2] + 1, 
				chain_matrix[x][y].str_1.substr(stacks[i][0] + stacks[i][2], stacks[x][1] - stacks[i][0] - stacks[i][2] + 1));
			str_chain_y = chain_matrix[x][y].str_2;
			int energy_new = chain_matrix[x][y].sum_free_energy + energy_model_i.Multi_closing + stacks[i][3]
				+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i])
				+ SHAPE_constraint.evaluate_pseudo_energy_exclude(stacks[i][0], stacks[i][1], stacks[i][0] + stacks[i][2], 
					stacks[x][1], str_multi_i, str_chain_y, frequency);
			if(energy_new < i_multi.sum_free_energy)	{
				/*
				if(i == 38 && j == 57 && y == 43)	{
					cout << "!!!!!!!!!!!!!!!!!!!!" << endl;
					cout << "chain_matrix energy_1:	" << chain_matrix[x][y].free_energy_1 << endl;
					cout << "chain_matrix energy_2:	" << chain_matrix[x][y].free_energy_2 << endl;
					cout << "loop energy:	" << energy_model_i.Multi_closing << endl;
					cout << "stack energy:	" << stacks[i][3] << endl;
					cout << "optimistic energy:	" << SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i]) << endl;
					cout << "constraint energh:	" << SHAPE_constraint.pseudo_energy_stack_constrained(stacks[i], str_chain_y, frequency) << endl;
					cout << decompress_structure(0, chain_matrix[x][y]) << endl;
					cout << decompress_structure(1, chain_matrix[x][y]) << endl;
					cout << "!!!!!!!!!!!!!!!!!!" << endl;
				}
				*/
				i_multi.free_energy_1 = energy_new - chain_matrix[x][y].free_energy_2;
				i_multi.free_energy_2 = chain_matrix[x][y].free_energy_2;
				i_multi.sum_free_energy = energy_new;
				i_multi.str_begin_1 = stacks[i][0];
				i_multi.str_end_1 = stacks[i][1];
				i_multi.str_stacks_1 = chain_matrix[x][y].str_stacks_1;
				i_multi.str_stacks_1.resize(i_multi.str_stacks_1.size() + 1);
				i_multi.str_stacks_1[i_multi.str_stacks_1.size() - 1] = i;
				i_multi.str_begin_2 = chain_matrix[x][y].str_begin_2;
				i_multi.str_end_2 = chain_matrix[x][y].str_end_2;
				i_multi.str_stacks_2 = chain_matrix[x][y].str_stacks_2;
			}
		}
		//if(i == 0 && j == 2)	{
		//	cout << "i_multi:	" << endl;
		//}
		
		
		if(y_constrained.sum_free_energy <= y_boundary.sum_free_energy &&
			y_constrained.sum_free_energy <= i_hairpin.sum_free_energy && 
			y_constrained.sum_free_energy <= i_internal.sum_free_energy && 
			y_constrained.sum_free_energy <= i_multi.sum_free_energy)	{
			/*
			if(i == 74 && j == 80 && y == 53)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << y << endl;
				cout << "chainmatrix iloop:	y_constrained:	" << y_constrained.free_energy_1 << "	" << y_constrained.free_energy_2 << endl;
			}
			*/
			copy_folding_matrix(y_constrained, chainloop_matrix_iloop[i][j][y]);
			
		}	else if(y_boundary.sum_free_energy <= y_constrained.sum_free_energy &&
			y_boundary.sum_free_energy <= i_hairpin.sum_free_energy && 
			y_boundary.sum_free_energy <= i_internal.sum_free_energy && 
			y_boundary.sum_free_energy <= i_multi.sum_free_energy)	{
			
			/*
			if(i == 74 && j == 80 && y == 53)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << y << endl;
				cout << "chainmatrix iloop:	y_boundary:	" << y_boundary.free_energy_1 << "	" << y_boundary.free_energy_2 << endl;
			}
			*/
			copy_folding_matrix(y_boundary, chainloop_matrix_iloop[i][j][y]);
			
		}	else if(i_hairpin.sum_free_energy <= y_constrained.sum_free_energy &&
			i_hairpin.sum_free_energy <= y_boundary.sum_free_energy && 
			i_hairpin.sum_free_energy <= i_internal.sum_free_energy && 
			i_hairpin.sum_free_energy <= i_multi.sum_free_energy)	{
			/*
			if(i == 74 && j == 80 && y == 53)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << y << endl;
				cout << "chainmatrix iloop:	i_hairpin:	" << i_hairpin.free_energy_1 << "	" << i_hairpin.free_energy_2 << endl;
			}
			*/
			copy_folding_matrix(i_hairpin, chainloop_matrix_iloop[i][j][y]);
		
		}	else if(i_internal.sum_free_energy <= y_constrained.sum_free_energy &&
			i_internal.sum_free_energy <= y_boundary.sum_free_energy && 
			i_internal.sum_free_energy <= i_hairpin.sum_free_energy && 
			i_internal.sum_free_energy <= i_multi.sum_free_energy)	{
			
			/*
			if(i == 74 && j == 80 && y == 53)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << y << endl;
				cout << "chainmatrix iloop:	i_internal:	" << i_internal.free_energy_1 << "	" << i_internal.free_energy_2 << endl;
			}
			*/
			copy_folding_matrix(i_internal, chainloop_matrix_iloop[i][j][y]);
			
		}	else if(i_multi.sum_free_energy <= y_constrained.sum_free_energy &&
			i_multi.sum_free_energy <= y_boundary.sum_free_energy && 
			i_multi.sum_free_energy <= i_hairpin.sum_free_energy && 
			i_multi.sum_free_energy <= i_internal.sum_free_energy)	{
			
			/*
			if(i == 74 && j == 80 && y == 53)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << y << endl;
				cout << "chainmatrix iloop:	i_multi:	" << i_multi.free_energy_1 << "	" << i_multi.free_energy_2 << endl;
			}
			*/
			copy_folding_matrix(i_multi, chainloop_matrix_iloop[i][j][y]);
		
		}
		chainloop_matrix_iloop[i][j][y].str_1 = decompress_structure(0, chainloop_matrix_iloop[i][j][y]);
		chainloop_matrix_iloop[i][j][y].str_2 = decompress_structure(1, chainloop_matrix_iloop[i][j][y]);
		/*
		if(i == 74 && j == 80 && y == 53)	{
			cout << "^^^^^^^^^^^^^^^^^" << endl;
			cout << i << "	" << j << "	" << y << endl;
			cout << "energy_1:	" << chainloop_matrix_iloop[i][j][y].free_energy_1 << endl;
			cout << decompress_structure(0, chainloop_matrix_iloop[i][j][y]) << endl;
			cout << "energy_2:	" << chainloop_matrix_iloop[i][j][y].free_energy_2 << endl;
			cout << decompress_structure(1, chainloop_matrix_iloop[i][j][y]) << endl;
		}
		*/
	}	// endfor
	
	return;
}

void bi_fold_stack::fill_chainloop_matrix_jloop(int i, int j, vector< vector<folding_matrix> > chain_matrix)	{
	
	int k, l, m, n;
	energy_parameters energy_model_i(sequence);
	energy_parameters energy_model_j(sequence);
	if(i == stacks.size() - 1)	{
		energy_model_i.set_pseudo_stack_energy();
	}
	if(j == stacks.size() - 1)	{
		energy_model_j.set_pseudo_stack_energy();
	}
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	
	for(k = num_enclosing[i] - 1; k >= 0; -- k)	{
		int x = enclosing[i][k];
		/* x constrained */
		folding_matrix x_constrained;
		x_constrained.sum_free_energy = LARGENUM;
		for(m = num_Fset[x] - 1; m >= 0; -- m)	{
			int xprev = Fset[x][m];
			if(stacks[xprev][0] >= stacks[i][0] + stacks[i][2])	{
				folding_matrix traverse_Fset;
				optimal_configuration(0, x, chainloop_matrix_jloop[i][j][xprev], traverse_Fset);
				string str_x = traverse_Fset.str_1;
				string str_y = chainloop_matrix_jloop[i][j][xprev].str_2;
				int constraint_energy = SHAPE_constraint.evaluate_pseudo_energy(stacks[xprev][1] + 1, stacks[x][1], str_x, str_y, frequency);
				if(traverse_Fset.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy < 0 && 
					chainloop_matrix_jloop[i][j][xprev].sum_free_energy + traverse_Fset.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy 
					< x_constrained.sum_free_energy)	{
					/* x taking the configuration as indicated by traverse_Fset */
					x_constrained.free_energy_1 = chainloop_matrix_jloop[i][j][xprev].free_energy_1
						+ traverse_Fset.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy;
					x_constrained.free_energy_2 = chainloop_matrix_jloop[i][j][xprev].free_energy_2;
					x_constrained.sum_free_energy = x_constrained.free_energy_1 + x_constrained.free_energy_2;
					/* fillup structures */
					x_constrained.str_begin_2 = chainloop_matrix_jloop[i][j][xprev].str_begin_2;
					x_constrained.str_end_2 = chainloop_matrix_jloop[i][j][xprev].str_end_2;
					x_constrained.str_stacks_2 = chainloop_matrix_jloop[i][j][xprev].str_stacks_2;
					string str_new = chainloop_matrix_jloop[i][j][xprev].str_1;
					str_new.replace(stacks[xprev][1] + 1, stacks[x][0] - stacks[xprev][1] - 1,
						string(stacks[x][0] - stacks[xprev][1] - 1, '.'));
					str_new.replace(stacks[x][0], stacks[x][1] - stacks[x][0] + 1, 
						traverse_Fset.str_1.substr(stacks[x][0], stacks[x][1] - stacks[x][0] + 1));
					compress_structure(0, str_new, x_constrained);
				}	else if(traverse_Fset.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy >= 0 && 
					chainloop_matrix_jloop[i][j][xprev].sum_free_energy < x_constrained.sum_free_energy)	{
					/* x taking no structure */
					x_constrained.free_energy_1 = chainloop_matrix_jloop[i][j][xprev].free_energy_1;
					x_constrained.free_energy_2 = chainloop_matrix_jloop[i][j][xprev].free_energy_2;
					x_constrained.sum_free_energy = chainloop_matrix_jloop[i][j][xprev].sum_free_energy;
					/* fillup structures */
					x_constrained.str_begin_1 = chainloop_matrix_jloop[i][j][xprev].str_begin_1;
					x_constrained.str_end_1 = stacks[x][1];
					x_constrained.str_stacks_1 = chainloop_matrix_jloop[i][j][xprev].str_stacks_1;
					x_constrained.str_begin_2 = chainloop_matrix_jloop[i][j][xprev].str_begin_2;
					x_constrained.str_end_2 = chainloop_matrix_jloop[i][j][xprev].str_end_2;
					x_constrained.str_stacks_2 = chainloop_matrix_jloop[i][j][xprev].str_stacks_2;
				}
				/*
				if(i == 57 && j == 36 && x == 43)	{
					cout << "&&&&&&&&&&&&&&&&&" << endl;
					cout << xprev << endl;
					cout << "optimal configuration energy:	" << traverse_Fset.free_energy_1 << endl;
					cout << "loop energy:	" << energy_model.Multi_new_branch << endl;
					cout << "constraint energy:	" << constraint_energy << endl;
					cout << x_constrained.free_energy_1 << "	" << x_constrained.free_energy_2 << endl;
					cout << decompress_structure(0, x_constrained) << endl;
					cout << decompress_structure(1, x_constrained) << endl;
				}
				*/
			}
		}
		
		
		/* x boundary */
		folding_matrix x_boundary;
		folding_matrix traverse_boundary_x;
		optimal_configuration(0, x, unconstrained_matrix[j], traverse_boundary_x);
		string str_x = traverse_boundary_x.str_1;
		string str_y = unconstrained_matrix[j].str_2;
		int constraint_energy = SHAPE_constraint.evaluate_pseudo_energy(stacks[i][0] + stacks[i][2], stacks[x][1], str_x, str_y, frequency);
		if(traverse_boundary_x.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy < 0)	{
			x_boundary.free_energy_1 = traverse_boundary_x.free_energy_1 + energy_model_i.Multi_new_branch + constraint_energy;
			x_boundary.free_energy_2 = unconstrained_matrix[j].free_energy_1;
			x_boundary.sum_free_energy = x_boundary.free_energy_1 + x_boundary.free_energy_2;
			x_boundary.str_begin_1 = stacks[i][0] + stacks[i][2];
			x_boundary.str_end_1 = stacks[x][1];
			x_boundary.str_stacks_1 = traverse_boundary_x.str_stacks_1;
			x_boundary.str_begin_2 = unconstrained_matrix[j].str_begin_1;
			x_boundary.str_end_2 = unconstrained_matrix[j].str_end_1;
			x_boundary.str_stacks_2 = unconstrained_matrix[j].str_stacks_1;
		}	else	{
			x_boundary.free_energy_1 = 0;
			x_boundary.free_energy_2 = unconstrained_matrix[j].free_energy_1;
			x_boundary.sum_free_energy = x_boundary.free_energy_1 + x_boundary.free_energy_2;
			x_boundary.str_begin_1 = stacks[i][0] + stacks[i][2];
			x_boundary.str_end_1 = stacks[x][1];
			x_boundary.str_begin_2 = unconstrained_matrix[j].str_begin_1 ;
			x_boundary.str_end_2 = unconstrained_matrix[j].str_end_1 ;
			x_boundary.str_stacks_2 = unconstrained_matrix[j].str_stacks_1;
		}
		/*
		if(i == 57 && j == 14 && x == 7)	{
			cout << "&&&&&&&&&&&&&&&&&" << endl;
			cout << "optimal configuration energy:	" << traverse_boundary_x.free_energy_1 << endl;
			cout << "loop energy:	" << energy_model.Multi_new_branch << endl;
			cout << "constraint energy:	" << constraint_energy << endl;
			cout << x_boundary.free_energy_1 << "	" << x_boundary.free_energy_2 << endl;
			cout << decompress_structure(0, x_boundary) << endl;
			cout << decompress_structure(1, x_boundary) << endl;
		}
		*/
		//if(i == 2 && j == 8 && x == 1)	{
		//	cout << "x_boundary" << endl;
		//	cout << decompress_structure(0, y_boundary) << endl;
		//}
		
			
		/* j as hairpin loop */
		folding_matrix j_hairpin;
		string str_chain_x = unconstrained_multi_matrix[i][x].str_1;
		string str_hairpin_j(sequence.length(), 'x');
		str_hairpin_j.replace(stacks[j][0], stacks[j][1] - stacks[j][0] + 1, string(stacks[j][1] - stacks[j][0] + 1, '.'));
		append_stack(stacks[j], str_hairpin_j);
		j_hairpin.free_energy_1 = unconstrained_multi_matrix[i][x].free_energy_1;
		j_hairpin.free_energy_2 = stacks[j][3] + energy_model_j.FE_hairpin(stacks[j][0] + stacks[j][2] - 1, stacks[j][1] - stacks[j][2] + 1) 
			+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[j])
			+ SHAPE_constraint.evaluate_pseudo_energy(stacks[j][0], stacks[j][1], str_hairpin_j, str_chain_x, 100 - frequency);
		j_hairpin.sum_free_energy = j_hairpin.free_energy_1 + j_hairpin.free_energy_2;
		j_hairpin.str_begin_1 = unconstrained_multi_matrix[i][x].str_begin_1;
		j_hairpin.str_end_1 = unconstrained_multi_matrix[i][x].str_end_1;
		j_hairpin.str_stacks_1 = unconstrained_multi_matrix[i][x].str_stacks_1;
		j_hairpin.str_begin_2 = stacks[j][0];
		j_hairpin.str_end_2 = stacks[j][1];
		j_hairpin.str_stacks_2.resize(1);
		j_hairpin.str_stacks_2[0] = j;
		
		//if(i == 2 && j == 8 && x == 1)	{
		//	cout << "j_hairpin" << endl;
		//	//cout << decompress_structure(0, i_hairpin) << endl;
		//}
		
		
		/* i as internal loop */
		folding_matrix j_internal;
		j_internal.sum_free_energy = LARGENUM;
		for(n = num_internal[j] - 1; n >= 0; -- n)	{
			int jen = internal[j][n];
			str_chain_x = chainloop_matrix_jloop[i][jen][x].str_1;
			//if(i == 2 && j == 8 && x == 1)	{
			//	cout << jen << endl;
			//	cout << str_chain_x << endl;
			//}
			string str_internal_j(sequence.length(), 'x');
			str_internal_j.replace(stacks[j][0], stacks[j][1] - stacks[j][0] + 1, string(stacks[j][1] - stacks[j][0] + 1, '.'));
			append_stack(stacks[j], str_internal_j);
			str_internal_j.replace(stacks[jen][0], stacks[jen][1] - stacks[jen][0] + 1, 
				chainloop_matrix_jloop[i][jen][x].str_2.substr(stacks[jen][0], stacks[jen][1] - stacks[jen][0] + 1));
			int energy_new = chainloop_matrix_jloop[i][jen][x].sum_free_energy + stacks[j][3]
				+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[j])
				+ energy_model_j.FE_loop(stacks[j][0] + stacks[j][2] - 1, stacks[j][1] - stacks[j][2] + 1, stacks[jen][1], stacks[jen][0])
				+ SHAPE_constraint.evaluate_pseudo_energy_exclude(stacks[j][0], stacks[j][1], stacks[jen][0], stacks[jen][1],
					str_internal_j, str_chain_x, 100 - frequency);
			if(energy_new < j_internal.sum_free_energy)	{
				j_internal.free_energy_1 = chainloop_matrix_jloop[i][jen][x].free_energy_1;
				j_internal.free_energy_2 = energy_new - chainloop_matrix_jloop[i][jen][x].free_energy_1;
				j_internal.sum_free_energy = energy_new;
				j_internal.str_begin_1 = chainloop_matrix_jloop[i][jen][x].str_begin_1;
				j_internal.str_end_1 = chainloop_matrix_jloop[i][jen][x].str_end_1;
				j_internal.str_stacks_1 = chainloop_matrix_jloop[i][jen][x].str_stacks_1;
				j_internal.str_begin_2 = stacks[j][0];
				j_internal.str_end_2 = stacks[j][1];
				j_internal.str_stacks_2 = chainloop_matrix_jloop[i][jen][x].str_stacks_2;
				j_internal.str_stacks_2.resize(j_internal.str_stacks_2.size() + 1);
				j_internal.str_stacks_2[j_internal.str_stacks_2.size() - 1] = j;
			}
			/*
			if(i == 57 && j == 14 && x == 7)	{
				cout << "&&&&&&&&&&&&&&&&&" << endl;
				cout << jen << endl;
				cout << "stack energy:	" << stacks[j][3] << endl;
				cout << "loop energy:	" << energy_model.FE_loop(stacks[j][0] + stacks[j][2] - 1, stacks[j][1] - stacks[j][2] + 1, stacks[jen][1], stacks[jen][0]) << endl;
				cout << "optimistic energy:	" << SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[j]) << endl;
				cout << "constraint energy:	" << SHAPE_constraint.pseudo_energy_stack_constrained(stacks[j], str_chain_x, 100 - frequency) << endl;
				cout << j_internal.free_energy_1 << "	" << j_internal.free_energy_2 << endl;
				cout << decompress_structure(0, j_internal) << endl;
				cout << decompress_structure(1, j_internal) << endl;
			}
			*/
			
		}
		
		
		
		/* i as multi loop */
		folding_matrix j_multi;
		j_multi.sum_free_energy = LARGENUM;
		for(n = num_FPset[j] - 1; n >= 0; -- n)	{
			int y = FPset[j][n];
			str_chain_x = chain_matrix[x][y].str_1;
			string str_multi_j(sequence.length(), 'x');
			str_multi_j.replace(stacks[j][0], stacks[j][1] - stacks[j][0] + 1, string(stacks[j][1] - stacks[j][0] + 1, '.'));
			append_stack(stacks[j], str_multi_j);
			str_multi_j.replace(stacks[j][0] + stacks[j][2], stacks[y][1] - stacks[j][0] - stacks[j][2] + 1, 
				chain_matrix[x][y].str_2.substr(stacks[j][0] + stacks[j][2], stacks[y][1] - stacks[j][0] - stacks[j][2] + 1));
			int energy_new = chain_matrix[x][y].sum_free_energy + energy_model_j.Multi_closing + stacks[j][3]
				+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[j])
				+ SHAPE_constraint.evaluate_pseudo_energy_exclude(stacks[j][0], stacks[j][1], stacks[j][0] + stacks[j][2], stacks[y][1], 
					str_multi_j, str_chain_x, 100 - frequency);
			if(energy_new < j_multi.sum_free_energy)	{
				//if(i == 18 && j == 18 && y == 11)	{
				//	cout << "chain_matrix" << endl;
				//	cout << decompress_structure(0, chain_matrix[x][y]) << endl;
				//}
				j_multi.free_energy_1 = chain_matrix[x][y].free_energy_1;
				j_multi.free_energy_2 = energy_new - chain_matrix[x][y].free_energy_1;
				j_multi.sum_free_energy = energy_new;
				j_multi.str_begin_1 = chain_matrix[x][y].str_begin_1;
				j_multi.str_end_1 = chain_matrix[x][y].str_end_1;
				j_multi.str_stacks_1 = chain_matrix[x][y].str_stacks_1;
				j_multi.str_begin_2 = stacks[j][0];
				j_multi.str_end_2 = stacks[j][1];
				j_multi.str_stacks_2 = chain_matrix[x][y].str_stacks_2;
				j_multi.str_stacks_2.resize(j_multi.str_stacks_2.size() + 1);
				j_multi.str_stacks_2[j_multi.str_stacks_2.size() - 1] = j;
			}
		}
		//if(i == 2 && j == 8 && x == 1)	{
		//	cout << "j_multi" << endl;
		//	//cout << decompress_structure(1, j_multi) << endl;
		//}
		
		
		if(x_constrained.sum_free_energy <= x_boundary.sum_free_energy &&
			x_constrained.sum_free_energy <= j_hairpin.sum_free_energy && 
			x_constrained.sum_free_energy <= j_internal.sum_free_energy && 
			x_constrained.sum_free_energy <= j_multi.sum_free_energy)	{
			/*
			if(i == 57 && j == 57 && x == 43)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << x << endl;
				cout << "chainmatrix jloop:	x_constrained" << endl;
			}
			*/
			copy_folding_matrix(x_constrained, chainloop_matrix_jloop[i][j][x]);
			
		}	else if(x_boundary.sum_free_energy <= x_constrained.sum_free_energy &&
			x_boundary.sum_free_energy <= j_hairpin.sum_free_energy && 
			x_boundary.sum_free_energy <= j_internal.sum_free_energy && 
			x_boundary.sum_free_energy <= j_multi.sum_free_energy)	{
			
			/*
			if(i == 57 && j == 57 && x == 43)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << x << endl;
				cout << "chainmatrix jloop:	x_boundary" << endl;
			}
			*/
			copy_folding_matrix(x_boundary, chainloop_matrix_jloop[i][j][x]);
			
		}	else if(j_hairpin.sum_free_energy <= x_constrained.sum_free_energy &&
			j_hairpin.sum_free_energy <= x_boundary.sum_free_energy && 
			j_hairpin.sum_free_energy <= j_internal.sum_free_energy && 
			j_hairpin.sum_free_energy <= j_multi.sum_free_energy)	{
			
			/*
			if(i == 57 && j == 57 && x == 43)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << x << endl;
				cout << "chainmatrix jloop:	j_hairpin" << endl;
			}
			*/
			copy_folding_matrix(j_hairpin, chainloop_matrix_jloop[i][j][x]);
		
		}	else if(j_internal.sum_free_energy <= x_constrained.sum_free_energy &&
			j_internal.sum_free_energy <= x_boundary.sum_free_energy && 
			j_internal.sum_free_energy <= j_hairpin.sum_free_energy && 
			j_internal.sum_free_energy <= j_multi.sum_free_energy)	{
			
			/*
			if(i == 57 && j == 57 && x == 43)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << x << endl;
				cout << "chainmatrix jloop:	j_internal" << endl;
			}
			*/
			copy_folding_matrix(j_internal, chainloop_matrix_jloop[i][j][x]);
			
		}	else if(j_multi.sum_free_energy <= x_constrained.sum_free_energy &&
			j_multi.sum_free_energy <= x_boundary.sum_free_energy && 
			j_multi.sum_free_energy <= j_hairpin.sum_free_energy && 
			j_multi.sum_free_energy <= j_internal.sum_free_energy)	{
			
			
			/*
			if(i == 57 && j == 57 && x == 43)	{
				cout << "************" << endl;
				cout << i << "	" << j << "	" << x << endl;
				cout << "chainmatrix jloop:	j_multi" << endl;
			}
			*/
			
			copy_folding_matrix(j_multi, chainloop_matrix_jloop[i][j][x]);
		
		}
		chainloop_matrix_jloop[i][j][x].str_1 = decompress_structure(0, chainloop_matrix_jloop[i][j][x]);
		chainloop_matrix_jloop[i][j][x].str_2 = decompress_structure(1, chainloop_matrix_jloop[i][j][x]);
		
		/*
		if(i == 57 && j == 57 && x == 43)	{
			cout << "^^^^^^^^^^^^^^^^^" << endl;
			cout << i << "	" << j << "	" << x << endl;
			cout << "energy_1:	" << chainloop_matrix_jloop[i][j][x].free_energy_1 << endl;
			cout << decompress_structure(0, chainloop_matrix_jloop[i][j][x]) << endl;
			cout << "energy_2:	" << chainloop_matrix_jloop[i][j][x].free_energy_2 << endl;
			cout << decompress_structure(1, chainloop_matrix_jloop[i][j][x]) << endl;
		}
		*/
	}	// endfor
	
	return;
}

void bi_fold_stack::fill_unconstrained_chain_matrix(int i, int j)	{

	int k, l, m, n;
	for(k = num_enclosing[i] - 1; k >= 0; -- k)	{
		int x = enclosing[i][k];
		chainloop_matrix_jloop[i][j][x].free_energy_1 = unconstrained_multi_matrix[i][x].free_energy_1;
		chainloop_matrix_jloop[i][j][x].free_energy_2 = unconstrained_matrix[j].free_energy_1;
		chainloop_matrix_jloop[i][j][x].sum_free_energy = chainloop_matrix_jloop[i][j][x].free_energy_1 + chainloop_matrix_jloop[i][j][x].free_energy_2;
		chainloop_matrix_jloop[i][j][x].str_begin_1 = unconstrained_multi_matrix[i][x].str_begin_1;
		chainloop_matrix_jloop[i][j][x].str_end_1 = unconstrained_multi_matrix[i][x].str_end_1;
		chainloop_matrix_jloop[i][j][x].str_stacks_1 = unconstrained_multi_matrix[i][x].str_stacks_1;
		chainloop_matrix_jloop[i][j][x].str_1 = unconstrained_multi_matrix[i][x].str_1;
		chainloop_matrix_jloop[i][j][x].str_begin_2 = unconstrained_matrix[j].str_begin_1;
		chainloop_matrix_jloop[i][j][x].str_end_2 = unconstrained_matrix[j].str_end_1;
		chainloop_matrix_jloop[i][j][x].str_stacks_2 = unconstrained_matrix[j].str_stacks_1;
		chainloop_matrix_jloop[i][j][x].str_2 = unconstrained_matrix[j].str_1;
	}
	
	for(l = num_enclosing[j] - 1; l >= 0; -- l)	{
		int y = enclosing[j][l];
		chainloop_matrix_iloop[i][j][y].free_energy_1 = unconstrained_matrix[i].free_energy_1;
		chainloop_matrix_iloop[i][j][y].free_energy_2 = unconstrained_multi_matrix[j][y].free_energy_1;
		chainloop_matrix_iloop[i][j][y].sum_free_energy = chainloop_matrix_iloop[i][j][y].free_energy_1 + chainloop_matrix_iloop[i][j][y].free_energy_2;
		chainloop_matrix_iloop[i][j][y].str_begin_1 = unconstrained_matrix[i].str_begin_1;
		chainloop_matrix_iloop[i][j][y].str_end_1 = unconstrained_matrix[i].str_end_1;
		chainloop_matrix_iloop[i][j][y].str_stacks_1 = unconstrained_matrix[i].str_stacks_1;
		chainloop_matrix_iloop[i][j][y].str_1 = unconstrained_matrix[i].str_1;
		chainloop_matrix_iloop[i][j][y].str_begin_2 = unconstrained_multi_matrix[j][y].str_begin_1;
		chainloop_matrix_iloop[i][j][y].str_end_2 = unconstrained_multi_matrix[j][y].str_end_1;
		chainloop_matrix_iloop[i][j][y].str_stacks_2 = unconstrained_multi_matrix[j][y].str_stacks_1;
		chainloop_matrix_iloop[i][j][y].str_2 = unconstrained_multi_matrix[j][y].str_1;
	}
	return;
}

void bi_fold_stack::fold_multi(int i, int j, folding_matrix &record_multi)	{
	
	int k, l, m, n;
	energy_parameters energy_model_i(sequence);
	if(i == stacks.size() - 1)	{
		energy_model_i.set_pseudo_stack_energy();
	}
	energy_parameters energy_model_j(sequence);
	if(j == stacks.size() - 1)	{
		energy_model_j.set_pseudo_stack_energy();
	}
		
	shape_reactivity SHAPE_constraint(sequence, reactivity);	
	record_multi.sum_free_energy = LARGENUM;	
	for(k = num_FPset[i] - 1; k >= 0; -- k)	{
		int x = FPset[i][k];
		string str_loop_i(sequence.length(), 'x');
		str_loop_i.replace(stacks[i][0], stacks[i][1] - stacks[i][0] + 1, string(stacks[i][1] - stacks[i][0] + 1, '.'));
		append_stack(stacks[i], str_loop_i);
		str_loop_i.replace(stacks[i][0] + stacks[i][2], stacks[x][1] - stacks[i][0] - stacks[i][2] + 1, 
			chainloop_matrix_jloop[i][j][x].str_1.substr(stacks[i][0] + stacks[i][2], stacks[x][1] - stacks[i][0] - stacks[i][2] + 1));
		string str_loop_j = chainloop_matrix_jloop[i][j][x].str_2;
		int energy_new = chainloop_matrix_jloop[i][j][x].sum_free_energy + stacks[i][3] + energy_model_i.Multi_closing
			+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i])
			+ SHAPE_constraint.evaluate_pseudo_energy_exclude(stacks[i][0], stacks[i][1], stacks[i][0] + stacks[i][2], stacks[x][1],
				str_loop_i, str_loop_j, frequency);
		if(energy_new < record_multi.sum_free_energy)	{
			record_multi.free_energy_1 = energy_new - chainloop_matrix_jloop[i][j][x].free_energy_2;
			record_multi.free_energy_2 = chainloop_matrix_jloop[i][j][x].free_energy_2;
			record_multi.sum_free_energy = energy_new;
			record_multi.str_begin_1 = stacks[i][0];
			record_multi.str_end_1 = stacks[i][1];
			record_multi.str_stacks_1 = chainloop_matrix_jloop[i][j][x].str_stacks_1;
			record_multi.str_stacks_1.resize(record_multi.str_stacks_1.size() + 1);
			record_multi.str_stacks_1[record_multi.str_stacks_1.size() - 1] = i;
			record_multi.str_begin_2 = chainloop_matrix_jloop[i][j][x].str_begin_2;
			record_multi.str_end_2 = chainloop_matrix_jloop[i][j][x].str_end_2;
			record_multi.str_stacks_2 = chainloop_matrix_jloop[i][j][x].str_stacks_2;
			/*
			if(i == 57 && j == 57)	{
				cout << "in fold_multi jloop:	" << x << endl;
				cout << "energy_1:	" << chainloop_matrix_jloop[i][j][x].free_energy_1 << endl;
				cout << decompress_structure(0, chainloop_matrix_jloop[i][j][x]) << endl;
				cout << "energy_2:	" << chainloop_matrix_jloop[i][j][x].free_energy_2 << endl;
				cout << decompress_structure(1, chainloop_matrix_jloop[i][j][x]) << endl;
			}
			*/
		}
		
		
		
	}
	
	for(l = num_FPset[j] - 1; l >= 0; -- l)	{
		int y = FPset[j][l];
		string str_loop_j(sequence.length(), 'x');
		str_loop_j.replace(stacks[j][0], stacks[j][1] - stacks[j][0] + 1, string(stacks[j][1] - stacks[j][0] + 1, '.'));
		append_stack(stacks[j], str_loop_j);
		str_loop_j.replace(stacks[j][0] + stacks[j][2], stacks[y][1] - stacks[j][0] - stacks[j][2] + 1, 
			chainloop_matrix_iloop[i][j][y].str_2.substr(stacks[j][0] + stacks[j][2], stacks[y][1] - stacks[j][0] - stacks[j][2] + 1));
		
		string str_loop_i = chainloop_matrix_iloop[i][j][y].str_1;
		int energy_new = chainloop_matrix_iloop[i][j][y].sum_free_energy + stacks[j][3] + energy_model_j.Multi_closing
			+ SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[j])
			+ SHAPE_constraint.evaluate_pseudo_energy_exclude(stacks[j][0], stacks[j][1], stacks[j][0] + stacks[j][2], stacks[y][1],
				str_loop_j, str_loop_i, 100 - frequency);
		if(energy_new < record_multi.sum_free_energy)	{
			record_multi.free_energy_1 = chainloop_matrix_iloop[i][j][y].free_energy_1;
			record_multi.free_energy_2 = energy_new - chainloop_matrix_iloop[i][j][y].free_energy_1;
			record_multi.sum_free_energy = energy_new;
			record_multi.str_begin_1 = chainloop_matrix_iloop[i][j][y].str_begin_1;
			record_multi.str_end_1 = chainloop_matrix_iloop[i][j][y].str_end_1;
			record_multi.str_stacks_1 = chainloop_matrix_iloop[i][j][y].str_stacks_1;
			record_multi.str_begin_2 = stacks[j][0];
			record_multi.str_end_2 = stacks[j][1];
			record_multi.str_stacks_2 = chainloop_matrix_iloop[i][j][y].str_stacks_2;
			record_multi.str_stacks_2.resize(record_multi.str_stacks_2.size() + 1);
			record_multi.str_stacks_2[record_multi.str_stacks_2.size() - 1] = j;
			
			/*
			if(i == 57 && j == 57)	{
				cout << "in fold_multi iloop:	" << y << endl;
				cout << "energy_1:	" << chainloop_matrix_iloop[i][j][y].free_energy_1 << endl;
				cout << decompress_structure(0, chainloop_matrix_iloop[i][j][y]) << endl;
				cout << "energy_2:	" << chainloop_matrix_iloop[i][j][y].free_energy_2 << endl;
				cout << decompress_structure(1, chainloop_matrix_iloop[i][j][y]) << endl;
			}
			*/
		}
		
		
		
	}
	
	return;
}

void bi_fold_stack::copy_folding_matrix(folding_matrix from, folding_matrix &to)	{

	to.sum_free_energy = from.sum_free_energy;
	
	to.free_energy_1 = from.free_energy_1;
	to.str_begin_1 = from.str_begin_1;
	to.str_end_1 = from.str_end_1;
	to.str_stacks_1 = from.str_stacks_1;
	
	to.free_energy_2 = from.free_energy_2;
	to.str_begin_2 = from.str_begin_2;
	to.str_end_2 = from.str_end_2;
	to.str_stacks_2 = from.str_stacks_2;
	
	return;
}

void bi_fold_stack::fold_Turner(void)	{
	
	int i, j, k, l, m;
	/* fill unconstrained matrix */
	fill_unconstrained_matrix();
	energy_parameters energy_model(sequence);
	
	for(i = 0; i < stacks.size(); ++ i)	{
		for(j = 0; j < stacks.size(); ++ j)	{
			/* check whether the two stacks overlap */
			//cout << "**************************" << endl;
			//cout << i << "	" << j << endl;
			//cout << stacks[i][0] << "	" << stacks[i][1] << "	" << stacks[i][2] << endl;
			//cout << stacks[j][0] << "	" << stacks[j][1] << "	" << stacks[j][2] << endl;
			
			if(!(is_overlap(stacks[i], stacks[j])))	{
				fold_unconstrained(i, j, constrained_matrix[i][j]);
				//cout << decompress_structure(0, constrained_matrix[i][j]) << endl;
				//cout << decompress_structure(1, constrained_matrix[i][j]) << endl;
				fill_unconstrained_chain_matrix(i, j);
			}	else	{
				folding_matrix record_hairpin;
				fold_hairpin(i, j, record_hairpin);
				//cout << record_hairpin.free_energy_1 << endl;
				//cout << decompress_structure(0, record_hairpin) << endl;
				//cout << record_hairpin.free_energy_2 << endl;
				//cout << decompress_structure(1, record_hairpin) << endl;
				//cout << "****************" << endl;
				folding_matrix record_internal;
				fold_internal(i, j, record_internal);
				
				vector< vector<folding_matrix> > chain_matrix (stacks.size());
				for(k = 0; k < stacks.size(); ++ k)	{
					chain_matrix[k] = vector<folding_matrix> (stacks.size());
				}
				
				fill_chain_matrix(i, j, chain_matrix);
				/*
				for(k = num_enclosing[i] - 1; k >= 0; -- k)	{
					for(l = num_enclosing[j] - 1; l >= 0; -- l)	{
						int x = enclosing[i][k];
						int y = enclosing[j][l];
						cout << "*******************" << endl;
						cout << i << "	" << j <<  "	" << x << "	" << y << endl;
						cout << decompress_structure(0, chain_matrix[x][y]) << endl;
						cout << decompress_structure(1, chain_matrix[x][y]) << endl;
					}
				}
				*/
				
				fill_chainloop_matrix_iloop(i, j, chain_matrix);
				
				/*
				for(l = num_enclosing[j] - 1; l >= 0; -- l)	{
					int y = enclosing[j][l];
					cout << "*******************" << endl;
					cout << i << "	" << j <<  "	" << y << endl;
					cout << decompress_structure(0, chainloop_matrix_iloop[i][j][y]) << endl;
					cout << decompress_structure(1, chainloop_matrix_iloop[i][j][y]) << endl;
				}
				*/
				
				
				fill_chainloop_matrix_jloop(i, j, chain_matrix);
				
				/*
				for(k = num_enclosing[i] - 1; k >= 0; -- k)	{
					int x = enclosing[i][k];
					cout << "*******************" << endl;
					cout << i << "	" << j <<  "	" << x << endl;
					cout << decompress_structure(0, chainloop_matrix_jloop[i][j][x]) << endl;
					cout << decompress_structure(1, chainloop_matrix_jloop[i][j][x]) << endl;
				}
				*/
				
				folding_matrix record_multi;
				fold_multi(i, j, record_multi);
				
				/*
				cout << "**********************" << endl;
				cout << i << "	" << j << endl;
				cout << decompress_structure(0, record_multi) << endl;
				cout << decompress_structure(1, record_multi) << endl;
				*/
				
				//cout << record_hairpin.sum_free_energy << "	" << record_internal.sum_free_energy << endl;
				if(record_hairpin.sum_free_energy <= record_internal.sum_free_energy && 
					record_hairpin.sum_free_energy <= record_multi.sum_free_energy)	{
					
					copy_folding_matrix(record_hairpin, constrained_matrix[i][j]);
				
				}	else if(record_internal.sum_free_energy <= record_hairpin.sum_free_energy && 
					record_internal.sum_free_energy <= record_multi.sum_free_energy)	{
					
					copy_folding_matrix(record_internal, constrained_matrix[i][j]);
					
				}	else if(record_multi.sum_free_energy <= record_hairpin.sum_free_energy && 
					record_multi.sum_free_energy <= record_internal.sum_free_energy)	{
				
					copy_folding_matrix(record_multi, constrained_matrix[i][j]);
				
				}
				
				constrained_matrix[i][j].real_energy_1 = evaluate_energy_stack(constrained_matrix[i][j].str_stacks_1);
				constrained_matrix[i][j].real_energy_2 = evaluate_energy_stack(constrained_matrix[i][j].str_stacks_2);
				constrained_matrix[i][j].str_1 = decompress_structure(0, constrained_matrix[i][j]);
				constrained_matrix[i][j].str_2 = decompress_structure(1, constrained_matrix[i][j]);
				/*
				if(i == 57 && j == 57)	{
					cout << i << "	" << j << endl;
					cout << record_hairpin.sum_free_energy << "	" << record_internal.sum_free_energy << "	" << record_multi.sum_free_energy << endl;
					cout << constrained_matrix[i][j].free_energy_1 << endl;
					cout << decompress_structure(0, constrained_matrix[i][j]) << endl;
					cout << constrained_matrix[i][j].free_energy_2 << endl;
					cout << decompress_structure(1, constrained_matrix[i][j]) << endl;
					cout << "eveluated energy for 1:	" << evaluate_energy_stack(constrained_matrix[i][j].str_stacks_1) << endl;
					cout << "eveluated energy for 2:	" << evaluate_energy_stack(constrained_matrix[i][j].str_stacks_2) << endl;
					cout << "*************" << endl;
				}
				*/
			}	// endif
		}	// endfor
	}	// endfor
	/*
	cout << "unconstraint_energy:	" << unconstrained_matrix[14].free_energy_1 << endl;
	cout << "chainloop jloop:	" << endl;
	cout << chainloop_matrix_jloop[57][57][43].free_energy_1 << endl;
	cout << decompress_structure(0, chainloop_matrix_jloop[57][57][43]) << endl;
	cout << chainloop_matrix_jloop[57][57][43].free_energy_2 << endl;
	cout << decompress_structure(1, chainloop_matrix_jloop[57][57][43]) << endl;
	cout << "real energy:	" << evaluate_energy_stack(unconstrained_matrix[14].str_stacks_1) << endl; 
	*/
	return;
}

int bi_fold_stack::evaluate_energy_stack(vector<int> stack_in_str)	{
	energy_parameters energy_model(sequence);
	int energy = 0;
	/* sorting the stack ids in stack_in_str */
	int i, j;
	for(i = 0; i < stack_in_str.size() - 1; ++ i)	{
		for(j = i + 1; j < stack_in_str.size(); ++ j)	{
			if(stack_in_str[i] > stack_in_str[j])	{
				int temp = i;
				i = j;
				j = temp;
			}
		}
	}
	/* traverse each stack and compute corresponding energy */
	vector<int> outmost_stacks;
	for(i = 0; i < stack_in_str.size(); ++ i)	{
		int m = stack_in_str[i];
		for(j = 0; j < outmost_stacks.size(); ++ j)	{
			int n = outmost_stacks[j]; 
			if(stacks[m][0] + stacks[m][2] - 1 < stacks[n][0])	{
				break;
			}
		}
		
		if(j == outmost_stacks.size())	{
			/* adding a hairpin loop energy */
			energy += energy_model.FE_hairpin(stacks[m][0] + stacks[m][2] - 1, stacks[m][1] - stacks[m][2] + 1);
			energy += stacks[m][3];	
			/* the new stack covers no one, add a new multi loop */
			outmost_stacks.resize(outmost_stacks.size() + 1);
			outmost_stacks[outmost_stacks.size() - 1] = m;
		}	else if(j == outmost_stacks.size() - 1)	{
			/* adding internal/bulge loop energy */
			int x = outmost_stacks[outmost_stacks.size() - 1];
			energy += energy_model.FE_loop(stacks[m][0] + stacks[m][2] - 1, stacks[m][1] - stacks[m][2] + 1, stacks[x][1], stacks[x][0]);
			energy += stacks[m][3];
			/* the new stack covers the last one, replace the last loop */
			outmost_stacks[outmost_stacks.size() - 1] = m;
		}	else if(outmost_stacks.size() - j > 1)	{
			/* adding multi loop energy */
			energy += energy_model.Multi_closing;
			energy += energy_model.Multi_new_branch * (outmost_stacks.size() - j);
			energy += stacks[m][3];
			/* the new stack covers several other stacks, record the last */
			outmost_stacks.resize(j + 1);
			outmost_stacks[outmost_stacks.size() - 1] = m;
		}
	}
	return energy;
}
