#include "./inc/energy_parameters.h"

using namespace std;

energy_parameters::energy_parameters(string input_sequence)	{
		
	int i, j, k;
		
	sequence = input_sequence;
		
	/* assign single parameters */
	Internal_assymmetric_penalty = 60;
	Internal_nonCG_closing_penalty = 70;
	Multi_closing = 800;
	Multi_new_branch = -100;
		
	/* assign loop free energy */
	Loop_internal = vector<int> (6);
	Loop_internal[0] = LARGENUM; Loop_internal[1] = 50; Loop_internal[2] = 160;
	Loop_internal[3] = 110; Loop_internal[4] = 210; Loop_internal[5] = 190;
		
	Loop_bulge = vector<int> (6); 
	Loop_bulge[0] = 380; Loop_bulge[1] = 280; Loop_bulge[2] = 320;
	Loop_bulge[3] = 360; Loop_bulge[4] = 400; Loop_bulge[5] = 440;
		
	Loop_hairpin = vector<int> (9); 
	Loop_hairpin[0] = LARGENUM; Loop_hairpin[1] = LARGENUM; Loop_hairpin[2] = 540;
	Loop_hairpin[3] = 560; Loop_hairpin[4] = 570; Loop_hairpin[5] = 540;
	Loop_hairpin[6] = 600; Loop_hairpin[7] = 550; Loop_hairpin[8] = 640;
		
	Loop_tri_hairpin = vector<int> (2);
	/*
		CAACG	6.8	GUUAC	6.9
	*/
	Loop_tri_hairpin[0] = 680; Loop_tri_hairpin[1] = 690;
		
	Loop_tetra_hairpin = vector<int> (15);
	/*
		CUACGG	2.8	CUCCGG	2.7	CUUCGG	3.7	CCAAGG	3.3	CCCAGG	3.4	
		CCGAGG	3.5	CCUAGG	3.7	CCACGG	3.7	CCGCGG	3.6	CCUCGG	2.5	
		CUAAGG	3.6	CUCAGG	3.7	CUUAGG	3.5	CUGCGG	2.8	CAACGG	5.5
	*/
	Loop_tetra_hairpin[0] = 280;  Loop_tetra_hairpin[1] = 270;  Loop_tetra_hairpin[2] = 370;  Loop_tetra_hairpin[3] = 330;  Loop_tetra_hairpin[4] = 340;
	Loop_tetra_hairpin[5] = 350;  Loop_tetra_hairpin[6] = 370;  Loop_tetra_hairpin[7] = 370;  Loop_tetra_hairpin[8] = 360;  Loop_tetra_hairpin[9] = 250;
	Loop_tetra_hairpin[10] = 360; Loop_tetra_hairpin[11] = 370; Loop_tetra_hairpin[12] = 350; Loop_tetra_hairpin[13] = 280; Loop_tetra_hairpin[14] = 550;
		
	Loop_hexa_hairpin = vector<int> (4);
	/*
		ACAGUGCU	2.9	ACAGUGAU	3.6	ACAGUUCU	1.8	ACAGUACU	2.8
	*/
	Loop_hexa_hairpin[0] = 290; Loop_hexa_hairpin[1] = 360; Loop_hexa_hairpin[2] = 180; Loop_hexa_hairpin[3] = 280;
		
	/* assign stacking free energy */
	/* index: AU->0, CG->1, GC->2, GU->3, UA->4, UG->5 */
	stacking_pairs = vector< vector<int> > (6);
	for(i = 0; i < 6; ++ i)	{
		stacking_pairs[i] = vector<int> (6);
	}
	stacking_pairs[0][0] = -90;  stacking_pairs[0][1] = -220; stacking_pairs[0][2] = -210; 
	stacking_pairs[0][3] = -60;  stacking_pairs[0][4] = -110; stacking_pairs[0][5] = -140;
	stacking_pairs[1][0] = -210; stacking_pairs[1][1] = -330; stacking_pairs[1][2] = -240;
	stacking_pairs[1][3] = -140; stacking_pairs[1][4] = -210; stacking_pairs[1][5] = -210;
	stacking_pairs[2][3] = -150; stacking_pairs[2][4] = -220; stacking_pairs[2][5] = -250;
	stacking_pairs[3][0] = -130; stacking_pairs[3][1] = -250; stacking_pairs[3][2] = -210;
	stacking_pairs[3][3] = -50;  stacking_pairs[3][4] = -140; stacking_pairs[3][5] = -130;
	stacking_pairs[4][0] = -130; stacking_pairs[4][1] = -240; stacking_pairs[4][2] = -210;
	stacking_pairs[4][3] = -100; stacking_pairs[4][4] = -90;  stacking_pairs[4][5] = -130;
	stacking_pairs[5][0] = -100; stacking_pairs[5][1] = -150; stacking_pairs[5][2] = -140;
	stacking_pairs[5][3] = -30;  stacking_pairs[5][4] = -60;  stacking_pairs[5][5] = -50;
	
	terminal_mismatch = vector< vector< vector<int> > > (6);
	for(i = 0; i < 6; ++ i)	{
		terminal_mismatch[i] = vector< vector<int> > (4);
		for(j = 0; j < 4; ++ j)	{
			terminal_mismatch[i][j] = vector<int> (4);
		}
	}
	terminal_mismatch[0][0][0] = -80;  terminal_mismatch[0][0][1] = -100; terminal_mismatch[0][0][2] = -80; terminal_mismatch[0][0][3] = -100;
	terminal_mismatch[0][1][0] = -60;  terminal_mismatch[0][1][1] = -70;  terminal_mismatch[0][1][2] = -60; terminal_mismatch[0][1][3] = -70;
	terminal_mismatch[0][2][0] = -80;  terminal_mismatch[0][2][1] = -100; terminal_mismatch[0][2][2] = -80; terminal_mismatch[0][2][3] = -100;
	terminal_mismatch[0][3][0] = -60;  terminal_mismatch[0][3][1] = -80;  terminal_mismatch[0][3][2] = -60; terminal_mismatch[0][3][3] = -80;
	
	terminal_mismatch[1][0][0] = -150; terminal_mismatch[1][0][1] = -150; terminal_mismatch[1][0][2] = -140; terminal_mismatch[1][0][3] = -150;
	terminal_mismatch[1][1][0] = -100; terminal_mismatch[1][1][1] = -100; terminal_mismatch[1][1][2] = -110; terminal_mismatch[1][1][3] = -80;
	terminal_mismatch[1][2][0] = -140; terminal_mismatch[1][2][1] = -150; terminal_mismatch[1][2][2] = -160; terminal_mismatch[1][2][3] = -150;
	terminal_mismatch[1][3][0] = -100; terminal_mismatch[1][3][1] = -140; terminal_mismatch[1][3][2] = -100; terminal_mismatch[1][3][3] = -120;
		
	terminal_mismatch[2][0][0] = -110; terminal_mismatch[2][0][1] = -150; terminal_mismatch[2][0][2] = -130; terminal_mismatch[2][0][3] = -150;
	terminal_mismatch[2][1][0] = -110; terminal_mismatch[2][1][1] = -70;  terminal_mismatch[2][1][2] = -110; terminal_mismatch[2][1][3] = -50;
	terminal_mismatch[2][2][0] = -160; terminal_mismatch[2][2][1] = -150; terminal_mismatch[2][2][2] = -140; terminal_mismatch[2][2][3] = -150;
	terminal_mismatch[2][3][0] = -110; terminal_mismatch[2][3][1] = -100; terminal_mismatch[2][3][2] = -110; terminal_mismatch[2][3][3] = -70;
		
	terminal_mismatch[3][0][0] = -30;  terminal_mismatch[3][0][1] = -100; terminal_mismatch[3][0][2] = -80;  terminal_mismatch[3][0][3] = -100;
	terminal_mismatch[3][1][0] = -60;  terminal_mismatch[3][1][1] = -70;  terminal_mismatch[3][1][2] = -60;  terminal_mismatch[3][1][3] = -70;
	terminal_mismatch[3][2][0] = -60;  terminal_mismatch[3][2][1] = -100; terminal_mismatch[3][2][2] = -80;  terminal_mismatch[3][2][3] = -100;
	terminal_mismatch[3][3][0] = -60;  terminal_mismatch[3][3][1] = -80;  terminal_mismatch[3][3][2] = -60;  terminal_mismatch[3][3][3] = -60;
		
	terminal_mismatch[4][0][0] = -100; terminal_mismatch[4][0][1] = -80;  terminal_mismatch[4][0][2] = -110; terminal_mismatch[4][0][3] = -80;
	terminal_mismatch[4][1][0] = -70;  terminal_mismatch[4][1][1] = -60;  terminal_mismatch[4][1][2] = -70;  terminal_mismatch[4][1][3] = -50;
	terminal_mismatch[4][2][0] = -110; terminal_mismatch[4][2][1] = -80;  terminal_mismatch[4][2][2] = -120; terminal_mismatch[4][2][3] = -80;
	terminal_mismatch[4][3][0] = -70;  terminal_mismatch[4][3][1] = -60;  terminal_mismatch[4][3][2] = -70;  terminal_mismatch[4][3][3] = -50;
		
	terminal_mismatch[5][0][0] = -100; terminal_mismatch[5][0][1] = -80;  terminal_mismatch[5][0][2] = -110; terminal_mismatch[5][0][3] = -80;
	terminal_mismatch[5][1][0] = -70;  terminal_mismatch[5][1][1] = -60;  terminal_mismatch[5][1][2] = -70;  terminal_mismatch[5][1][3] = -50;
	terminal_mismatch[5][2][0] = -50;  terminal_mismatch[5][2][1] = -80;  terminal_mismatch[5][2][2] = -80;  terminal_mismatch[5][2][3] = -80;
	terminal_mismatch[5][3][0] = -70;  terminal_mismatch[5][3][1] = -60;  terminal_mismatch[5][3][2] = -70;  terminal_mismatch[5][3][3] = -50;
		
	Internal_GG = vector< vector<int> > (6);
	for(i = 0; i < 6; ++ i)	{
		Internal_GG[i] = vector<int> (6);
	}
	Internal_GG[0][0] = -70;  Internal_GG[0][1] = -140; Internal_GG[0][2] = -140; 
	Internal_GG[0][3] = -70;  Internal_GG[0][4] = -70;  Internal_GG[0][5] = -70;
	Internal_GG[1][0] = -140; Internal_GG[1][1] = -140; Internal_GG[1][2] = -220; 
	Internal_GG[1][3] = -140; Internal_GG[1][4] = -140; Internal_GG[1][5] = -140;
	Internal_GG[2][0] = -140; Internal_GG[2][1] = -220; Internal_GG[2][2] = -230; 
	Internal_GG[2][3] = -140; Internal_GG[2][4] = -140; Internal_GG[2][5] = -140;
	Internal_GG[3][0] = -70;  Internal_GG[3][1] = -140; Internal_GG[3][2] = -140; 
	Internal_GG[3][3] = -70;  Internal_GG[3][4] = -70;  Internal_GG[3][5] = -70;
	Internal_GG[4][0] = -70;  Internal_GG[4][1] = -140; Internal_GG[4][2] = -140; 
	Internal_GG[4][3] = -70;  Internal_GG[4][4] = -70;  Internal_GG[4][5] = -70;
	Internal_GG[5][0] = -70;  Internal_GG[5][1] = -140; Internal_GG[5][2] = -140; 
	Internal_GG[5][3] = -70;  Internal_GG[5][4] = -70;  Internal_GG[5][5] = -70;
		
	return;
}


energy_parameters::~energy_parameters()	{
	return;
}

void energy_parameters::set_pseudo_stack_energy(void)	{
	Multi_closing = 0;
	Multi_new_branch = 0;
	return;
}
	
void energy_parameters::recover_pseudo_stack_energy(void)	{
	Multi_closing = 900;
	Multi_new_branch = -150;
	return;
}

int energy_parameters::FE_loop(int fiveP_out, int threeP_out, int fiveP_in, int threeP_in)	{
	
	/* case of pseudo stack */
	if(fiveP_out == -1 && threeP_out == sequence.length())	{
		return 0;
	}
		
	/* case of a bulge loop */
	if(threeP_in - fiveP_out == 1 || threeP_out - fiveP_in == 1)	{
		int n;
		if(threeP_in - fiveP_out == 1)	{
			n = threeP_out - fiveP_in - 1;
		}	else	{
			n = threeP_in - fiveP_out - 1;
		}
		if(n <= 6 && n > 0)	{
			/* note that the index starts from 0, i.e Loop_bulge[0] stores the energy for bulge loop size 1 */
			return Loop_bulge[n - 1];
		}	else if(n > 6)	{
			return Loop_bulge[5] + (int) (1.08 * log(((double) n) / 6) * 100);
		}	else	{
			return LARGENUM;
		}
	}	
	/* case of internal loop */
	else if(threeP_in - fiveP_out > 1 && threeP_out - fiveP_in > 1)	{
		int n1, n2;
		n1 = threeP_in - fiveP_out - 1;
		n2 = threeP_out - fiveP_in - 1;
		/* special case for GG symmetric internal loop */
		if(n1 == 1 && n2 == 1 && sequence[fiveP_out + 1] == 'G' && sequence[threeP_out - 1] == 'G')	{
			return Internal_GG[index_pair(sequence[fiveP_out], sequence[threeP_out])][index_pair(sequence[fiveP_in], sequence[threeP_in])];	
		}
		int num_assymmetric_nuc = abs(n1 - n2);
		int num_nonCG_closing = 0;
		if(!((sequence[fiveP_out] == 'C' &&  sequence[threeP_out] == 'G') || (sequence[fiveP_out] == 'G' &&  sequence[threeP_out] == 'C'))) {
			++ num_nonCG_closing;
		}	
		if(!((sequence[fiveP_in] == 'C' &&  sequence[threeP_in] == 'G') || (sequence[fiveP_in] == 'G' &&  sequence[threeP_in] == 'C'))) {
			++ num_nonCG_closing;
		}

		int internal_init;
		if(n1 + n2 <= 6 && n1 > 0 && n2 > 0)	{
			internal_init = Loop_internal[n1 + n2 - 1];
		}	else if(n1 + n2 > 6 && n1 > 0 && n2 > 0)	{
			internal_init = Loop_internal[5] + (int) (1.08 * log(((double) n1 + n2) / 6) * 100);
		}	else	{
			internal_init = LARGENUM;
		}
		return internal_init + num_assymmetric_nuc * Internal_assymmetric_penalty + num_nonCG_closing * Internal_nonCG_closing_penalty; 
	}	else	{
		return LARGENUM;
	}
}

int energy_parameters::FE_hairpin(int fiveP, int threeP)	{
	
	/* case of pseudo stack */
	if(fiveP == -1 && threeP == sequence.length())	{
		return 0;
	}
	
	int n = threeP - fiveP - 1;
	if(n <= 9 && n >= 3)	{
		string the_hairpin = sequence.substr(fiveP, threeP - fiveP + 1);
		if(n == 3)	{
			/* triloop special case */
			if(the_hairpin == "CAACG")	{
				return Loop_tri_hairpin[0];
			}	else if("GUUAC")	{
				return Loop_tri_hairpin[1];
			}	else	{
				return Loop_hairpin[n - 1]; 
				//+ terminal_mismatch[index_pair(sequence[fiveP], sequence[threeP])][index_nuc(sequence[fiveP + 1])][index_nuc(sequence[threeP - 1])];
				}
		}	else if(n == 4)	{
			/* tetraloop special case */
			if(the_hairpin == "CUACGG")	{
				return Loop_tetra_hairpin[0];
			}	else if(the_hairpin == "CUCCGG")	{
				return Loop_tetra_hairpin[1];
			}	else if(the_hairpin == "CUUCGG")	{
				return Loop_tetra_hairpin[2];
			}	else if(the_hairpin == "CCAAGG")	{
				return Loop_tetra_hairpin[3];
			}	else if(the_hairpin == "CCCAGG")	{
				return Loop_tetra_hairpin[4];
			}	else if(the_hairpin == "CCGAGG")	{
				return Loop_tetra_hairpin[5];
			}	else if(the_hairpin == "CCUAGG")	{
				return Loop_tetra_hairpin[6];
			}	else if(the_hairpin == "CCACGG")	{
				return Loop_tetra_hairpin[7];
			}	else if(the_hairpin == "CCGCGG")	{
				return Loop_tetra_hairpin[8];
			}	else if(the_hairpin == "CCUCGG")	{
				return Loop_tetra_hairpin[9];
			}	else if(the_hairpin == "CUAAGG")	{
				return Loop_tetra_hairpin[10];
			}	else if(the_hairpin == "CUCAGG")	{
				return Loop_tetra_hairpin[11];
			}	else if(the_hairpin == "CUUAGG")	{
				return Loop_tetra_hairpin[12];
			}	else if(the_hairpin == "CUGCGG")	{
				return Loop_tetra_hairpin[13];
			}	else if(the_hairpin == "CAACGG")	{
				return Loop_tetra_hairpin[14];
			}	else	{
				return Loop_hairpin[n - 1]; 
				//+ terminal_mismatch[index_pair(sequence[fiveP], sequence[threeP])][index_nuc(sequence[fiveP + 1])][index_nuc(sequence[threeP - 1])];
			}	
		}	else if(n == 6)	{
			/* hexa special case */
			if(the_hairpin == "ACAGUGCU")	{
				return Loop_hexa_hairpin[0];
			}	else if(the_hairpin == "ACAGUGAU")	{
				return Loop_hexa_hairpin[1];
			}	else if(the_hairpin == "ACAGUUCU")	{
				return Loop_hexa_hairpin[2];
			}	else if(the_hairpin == "ACAGUACU")	{
				return Loop_hexa_hairpin[3];
			}	else	{
				return Loop_hairpin[n - 1]; 
				//+ terminal_mismatch[index_pair(sequence[fiveP], sequence[threeP])][index_nuc(sequence[fiveP + 1])][index_nuc(sequence[threeP - 1])];
			}
		}	else	{	
			return Loop_hairpin[n - 1]; 
				//+ terminal_mismatch[index_pair(sequence[fiveP], sequence[threeP])][index_nuc(sequence[fiveP + 1])][index_nuc(sequence[threeP - 1])];
		}
	}	else if(n > 9)	{
		return Loop_hairpin[8] + (int) (1.08 * log(((double) n) / 9) * 100);
			//+ terminal_mismatch[index_pair(sequence[fiveP], sequence[threeP])][index_nuc(sequence[fiveP + 1])][index_nuc(sequence[threeP - 1])];
	}	else	{
		return LARGENUM;
	}
}	

int energy_parameters::index_pair(char a, char b)	{
	a = tolower(a);
	b = tolower(b);
		
	if(a == 'a' && b == 'u')	{
		return 0;
	}	else if(a == 'c' && b == 'g')	{
		return 1;
	}	else if(a == 'g' && b == 'c')	{
		return 2;
	}	else if(a == 'g' && b == 'u')	{
		return 3;
	}	else if(a == 'u' && b == 'a')	{
		return 4;
	}	else if(a == 'u' && b == 'g')	{
		return 5;
	}	else	{
		return -1;
	}
}

int energy_parameters::index_nuc(char a)	{
	a = tolower(a);
		
	if(a == 'a')	{
		return 0;
	}	else if(a == 'c')	{
		return 1;
	}	else if(a == 'g')	{
		return 2;
	}	else if(a == 'u')	{
		return 3;
	}	else	{
		return -1;
	}
}

int energy_parameters::FE_stack(int fiveP, int threeP)	{
	
	/* case of pseudo stack */
	if(fiveP == -1 && threeP == sequence.length())	{
		return 0;
	}

	if(fiveP < threeP && fiveP >= 0 && threeP < sequence.length())	{
		int inner_index = index_pair(sequence[threeP - 1], sequence[fiveP + 1]);
		int outer_index = index_pair(sequence[threeP], sequence[fiveP]);
		if(inner_index >= 0 && outer_index >= 0)	{
			return stacking_pairs[inner_index][outer_index];
		}	else	{
			return LARGENUM;
		}
	}	else	{
		return LARGENUM;
	}
}
