#include <iostream>
#include <string>
#include <vector>

#include "single_fold_stack.h"

#define MIN_LOOP 3
#define REACTIVIT_WEIGHT 5
#define LARGENUM 999999

using namespace std;

struct folding_matrix	{
	/* the free energy of this partial folding */
	int sum_free_energy;
	int free_energy_1, free_energy_2, str_begin_1, str_end_1, str_begin_2, str_end_2;
	int real_energy_1, real_energy_2;
	vector<int> str_stacks_1, str_stacks_2;
	string str_1, str_2;
};

class bi_fold_stack	{
	public:
	string sequence;
	string SHAPE_file;
	/* here the frequency refers to the frequency of transcript 1 */
	int frequency;
	string structure_1, structure_2;
	int fold_energy;
	
	vector< vector<int> > stacks;
	vector< vector<int> > stack_hash;
	
	vector<int> num_internal;
	vector< vector<int> > internal;
	
	vector<int> num_enclosing;
	vector< vector<int> > enclosing;
	
	vector<int> num_Fset;
	vector< vector<int> > Fset;
	
	vector<int> num_FPset;
	vector< vector<int> > FPset;
	
	vector<folding_matrix> unconstrained_matrix;
	vector< vector<folding_matrix> > unconstrained_multi_matrix;
	vector< vector<folding_matrix> > constrained_matrix;
	
	vector< vector< vector<folding_matrix> > > chainloop_matrix_iloop;
	vector< vector< vector<folding_matrix> > > chainloop_matrix_jloop;
	
	vector<double> reactivity;
	
	int min_stack_length, min_stack_hydro_bonds;
	
	bi_fold_stack(string input_sequence, string input_SHAPE_file, int input_frequency, 
		int input_min_stack_length, int input_min_stack_hydro_bonds);
	~bi_fold_stack(void);
	
	int num_hydro_bonds(char a, char b);
	void fold_Turner(void);
	void trace_back_Turner(int i, int j);
	void build_stacks(void);
	
	folding_matrix* access_energy_matrix(int i1, int j1, int i2, int j2);
	folding_matrix* access_unconstrained_matrix(int i, int j);
	int access_partial_score_all(int s, int i1, int j1, int i2, int j2, int &type);
	
	void fill_unconstrained_matrix(void);
	
	/* the indeces i and j indicates the ith stack and jth stack */
	void fold_unconstrained(int i, int j, folding_matrix &record_unconstrained);
	
	void fold_hairpin(int i, int j, folding_matrix &record_hairpin);
	void fold_internal(int i, int j, folding_matrix &record_internal);
	void fold_multi(int i, int j, folding_matrix &record_multi);
	
	void fill_chain_matrix(int i, int j, vector< vector<folding_matrix> > &chain_matrix);
	void fill_chainloop_matrix_iloop(int i, int j, vector< vector<folding_matrix> > chain_matrix);
	void fill_chainloop_matrix_jloop(int i, int j, vector< vector<folding_matrix> > chain_matrix);
	void fill_unconstrained_chain_matrix(int i, int j);
	
	void copy_folding_matrix(folding_matrix from, folding_matrix &to);
	
	void fold_Nussinov_constrained_extend(int i1, int j1, int i2, int j2);
	
	int predict_stacks(int min_length, int min_hydro_bonds);
	void sort_stacks(void);
	/* if which == 0, assign first structure holder, otherwise assign second structure holder */
	void compress_structure(bool which, string s, folding_matrix &m);
	string decompress_structure(bool which, folding_matrix m);
	void append_stack(vector<int> stack, string &structure);
	bool is_overlap(vector<int> stack_1, vector<int> stack_2);
	void optimal_configuration(bool which, int i, folding_matrix constraint, folding_matrix &optimal);
	
	int evaluate_energy_stack(vector<int> stack_in_str);
	
};
