#include <iostream>
#include <string>
#include <stdlib.h>
#include <cstdlib>
#include <unistd.h>

#include "./inc/bi_fold_stack.h"
#include "./inc/read_sequence.h"

string sequence_file, reactivity_file;
int min_stack_len = 4, min_stack_hydro_bonds = 8;
int frequency = 50;

void printUsage(void)	{
	cout << "Usage: MutualFold -i sequence_file -r reactivity_file [-l min_stack_len -b min_stack_hydro_bonds -f frequency]" << endl;
	cout << "sequence_file (-i):" << endl << "	" << "The file containning the sequence to be folded in FASTA format." << endl;
	cout << "reactivity_file (-r):" << endl << "	" << "The file containning the SHAPE reactivity for the RNA transcript." << endl;
	cout << "min_stack_len (-l):" << endl << "	" << "Minimum length for a stack to be considered significant. Default: 4" << endl;
	cout << "min_stack_hydro_bonds (-b):" << endl << "	" 
		<< "Minimum number of hydrogen bonds required for a stack to be considered significant. Default: 8" << endl;
	cout << "frequency (-f):" << endl << "	" << "Estimated frequency for the first transcript in percentage (between 1 to 99). Default: 50" << endl;
	return;
}

void initenv(int argc, char **argv)	{
	//initialize environment variable
	int opt = 0;
	while((opt=getopt(argc, argv, "i:r:l:b:f:h")) != -1) {
		switch(opt) {
			case 'i': {sequence_file = string(optarg); break;}
			case 'r': {reactivity_file = string(optarg); break;}
			case 'l': {min_stack_len = atoi(optarg); break;}
			case 'b': {min_stack_hydro_bonds = atoi(optarg); break;}
			case 'f': {frequency = atoi(optarg); break;}
			case 'h': {printUsage(); exit(0);}
			default: {printUsage(); exit(0);}
		}
	}
	if(sequence_file.empty() || reactivity_file.empty() || frequency < 1 || frequency > 99){
		printUsage();
		exit(0);
	}
	return;
}


int main(int argc, char **argv)	{
	
	initenv(argc, argv);
	read_sequence RNA_sequence(sequence_file);	
	
	/*
	shape_reactivity SHAPE_constraint(RNA_sequence.sequence, reactivity_file);
	cout << SHAPE_constraint.pseudo_energy_unpair_constrained(10, '.', 50) << endl;
	*/
	/*
	single_fold_stack constrained(RNA_sequence.sequence, reactivity_file, min_stack_len, min_stack_hydro_bonds);
	
	constrained.fold();
	constrained.traceback_stack(constrained.stacks.size() - 1);
	cout << constrained.structure << endl;
	constrained.clear_structure();
	
	constrained.fold_SHAPE();
	constrained.traceback_stack(constrained.stacks.size() - 1);
	cout << constrained.structure << endl;
	*/
	
	
	/*
	bi_fold_stack Mutual_constrained(RNA_sequence.sequence, reactivity_file, frequency, min_stack_len, min_stack_hydro_bonds);
	shape_reactivity SHAPE_constraint(RNA_sequence.sequence, reactivity_file);
	string str1("............(((((((.....)))))))(((((((((....)))))))))...(((((......)))))(((((.((((...........)))).)))))............");
	string cons("((((((((....(((((((.....)))))))(((((((((....)))))))))...(((((......)))))...)))))))).........((((((((.......))))))))");
	cout << "constraint energy 1:	" << SHAPE_constraint.evaluate_pseudo_energy(0, cons.length() - 1, str1, cons, frequency) << endl;
	folding_matrix check_energy;
	Mutual_constrained.compress_structure(0, str1, check_energy);
	Mutual_constrained.compress_structure(1, cons, check_energy);
	cout << "real energy 1:	" << Mutual_constrained.evaluate_energy_stack(check_energy.str_stacks_1) << endl;
	cout << "real energy 2:	" << Mutual_constrained.evaluate_energy_stack(check_energy.str_stacks_2) << endl;
	cout << "optimistic energy 1:	" << SHAPE_constraint.optimistic_pseudo_energy_structure(check_energy.str_stacks_1, Mutual_constrained.stacks) << endl;
	cout << "optimistic energy 2:	" << SHAPE_constraint.optimistic_pseudo_energy_structure(check_energy.str_stacks_2, Mutual_constrained.stacks) << endl;
	*/
	
	
	
	bi_fold_stack Mutual_constrained(RNA_sequence.sequence, reactivity_file, frequency, min_stack_len, min_stack_hydro_bonds);
	
	cout << "Num. of predicted stacks:	" << Mutual_constrained.stacks.size() << endl;
	Mutual_constrained.fold_Turner();

	int num_stack = Mutual_constrained.stacks.size();	
	
	int real_energy_1 = Mutual_constrained.evaluate_energy_stack(Mutual_constrained.constrained_matrix[num_stack - 1][num_stack - 1].str_stacks_1);
	int real_energy_2 = Mutual_constrained.evaluate_energy_stack(Mutual_constrained.constrained_matrix[num_stack - 1][num_stack - 1].str_stacks_2);
	
	
	cout << endl;
	cout << "*************************************************************" << endl;
	cout << "	Mutual constrained bi-folding results for: " << RNA_sequence.header << endl << endl;
	cout << "*************************************************************" << endl;
	
	cout << "Transcript 1:	" << endl;
	cout << "	Pseudo free energy:	" << Mutual_constrained.constrained_matrix[num_stack - 1][num_stack - 1].free_energy_1 / 100.0 << " kcal/mol"<< endl;
	//cout << "	Real free energy:	" << real_energy_1 / 100.0 << "	kcal/mol" << endl;
	cout << "	Frequency (as input):	" << frequency << "%" << endl;
	cout << "	Predicted structure:	" << endl;
	cout << "	" << RNA_sequence.sequence << endl;
	cout << "	" << Mutual_constrained.decompress_structure(0, Mutual_constrained.constrained_matrix[num_stack - 1][num_stack - 1]) << endl;
	cout << endl;
	
	cout << "Transcript 2:	" << endl;
	cout << "	Pseudo free energy:	" << Mutual_constrained.constrained_matrix[num_stack - 1][num_stack - 1].free_energy_2 / 100.0 << " kcal/mol"<< endl;
	//cout << "	Real free energy:	" << real_energy_2 / 100.0 << "	kcal/mol" << endl;
	cout << "	Frequency (as input):	" << 100 - frequency << "%" << endl;
	cout << "	Predicted structure:	" << endl;
	cout << "	" << RNA_sequence.sequence << endl;
	cout << "	" << Mutual_constrained.decompress_structure(1, Mutual_constrained.constrained_matrix[num_stack - 1][num_stack - 1]) << endl;
	cout << endl;
	
	
	return 0;
}
