#include "./inc/shape_reactivity.h"

using namespace std;

shape_reactivity::shape_reactivity(string input_sequence, vector<double> input_reactivity)	{
	
	slope = 2.6;
	intercept = -0.8;
	if(input_sequence.length() != input_reactivity.size())	{
		cout << "Error in assigning SHAPE reactivity!" << endl;
	}
	sequence = input_sequence;
	reactivity = input_reactivity;
	
	return;
}

shape_reactivity::shape_reactivity(string input_sequence, string shape_file_name)	{
		
	int i;
	slope = 2.6;
	intercept = -0.8;
	sequence = input_sequence;
	reactivity = vector<double> (sequence.length());
	char *file_name_holder = new char [shape_file_name.length() + 1];
	for(i = 0; i < shape_file_name.length(); ++ i)	{
		file_name_holder[i] = shape_file_name[i];
	}
	file_name_holder[shape_file_name.length()] = '\0';
	
	ifstream input_shape(file_name_holder, ifstream::in);
	if(!input_shape.good())	{
		cout << "Error in reading SHAPE file!" << endl;
	}
		
	delete [] file_name_holder;
	
	i = 0;
	while(input_shape.good())	{
		if(i >= sequence.length())	break;
		char a_line[100], nuc_reactivity[100];
		input_shape.getline(a_line, 100);
		int n = 0;
		while(a_line[n] != '\t')	{
			++ n;
		}
		//cout << input_shape.good() << endl;
		strcpy(nuc_reactivity, a_line + n + 1);
		reactivity[i] = atof(nuc_reactivity) >= 0 ? atof(nuc_reactivity) : 0;
		reactivity[i] = reactivity[i] >= 1.0 ? 1.0 : reactivity[i];
		//cout << reactivity[i] << endl;
		++ i;
	}
	input_shape.close();
	
	if(sequence.length() != reactivity.size())	{
		cout << sequence.length() << "	" << reactivity.size() << endl;
		cout << "Incompatible sequence and SHAPE profile!" << endl;
		exit(0);
	}
}

shape_reactivity::~shape_reactivity(void)	{
	return;
}

int shape_reactivity::pseudo_energy_pair(int i)	{
	if(i >= 0 && i < sequence.length())	{
		return (int) 100.0 * (slope * log(reactivity[i] + 1.0) + intercept);
	}	else	{
		return 0;
	}
}

int shape_reactivity::pseudo_energy_pair_constrained(int i, char constraint, int frequency)	{
	/* the frequency means the frequency of the current transcript */
	/* in another word, the frequency of the constraint is 100 - frequency */
	
	/* infer real reactivity */
	double df_frequency = ((double) frequency) / 100.0;
	double partial_reactivity;
	if(constraint == 'x')	{
		partial_reactivity = 0;
	}	else if(constraint == '.')	{
		partial_reactivity = (reactivity[i] - 1.0 * (1.0 - df_frequency)) / df_frequency;
	}	else	{
		partial_reactivity = (reactivity[i] - 0.0 * (1.0 - df_frequency)) / df_frequency;
	}
	/* check boundary */
	//cout << partial_reactivity << endl;
	partial_reactivity = fabs(partial_reactivity - 0.0);
	//cout << partial_reactivity << endl;
	if(i >= 0 && i < sequence.length())	{
		return (int) 100.0 * (slope * log(partial_reactivity + 1.0) + intercept) - optimistic_pseudo_energy_pair();
	}	else	{
		return 0;
	}
}

int shape_reactivity::pseudo_energy_unpair(int i)	{
	/* follow conventional definition of pseudo energy */
	return 0;
}

int shape_reactivity::pseudo_energy_unpair_constrained(int i, char constraint, int frequency)	{
	
	double df_frequency = ((double) frequency) / 100.0;
	double partial_reactivity;
	if(constraint == 'x')	{
		partial_reactivity = 1.0;
	}	else if(constraint == '.')	{
		partial_reactivity = 1.0;
	}	else	{
		partial_reactivity = (reactivity[i] - 0.0 * (1.0 - df_frequency)) / df_frequency;
	}
	/* check boundary */
	//cout << partial_reactivity << endl;
	partial_reactivity = fabs(partial_reactivity - 1.0);
	//cout << partial_reactivity << endl;
	if(i >= 0 && i < sequence.length())	{
		return (int) 100.0 * (slope * log(partial_reactivity + 1.0) + intercept) - optimistic_pseudo_energy_pair();
	}	else	{
		return 0;
	}
	
}

int shape_reactivity::pseudo_energy_stack(vector<int> stack)	{
		
	int energy = 0;
	int i;
	for(i = 0; i < stack[2]; ++ i)	{
		energy += pseudo_energy_pair(stack[0] + i);
		energy += pseudo_energy_pair(stack[1] - i);
	}
	return energy;
}

int shape_reactivity::pseudo_energy_stack_constrained(vector<int> stack, string constraint_str, int frequency)	{
		
	int energy = 0;
	int i;
	for(i = 0; i < stack[2]; ++ i)	{
		//cout << "&&&&&&&&&&&&&" << endl;
		//cout << "i:	" << i << endl;
		//cout << pseudo_energy_pair_constrained(stack[0] + i, constraint_str[stack[0] + i], frequency) << endl;
		//cout << pseudo_energy_pair_constrained(stack[1] - i, constraint_str[stack[1] - i], frequency) << endl;
		energy += pseudo_energy_pair_constrained(stack[0] + i, constraint_str[stack[0] + i], frequency);
		energy += pseudo_energy_pair_constrained(stack[1] - i, constraint_str[stack[1] - i], frequency);
	}
	return energy;
}

int shape_reactivity::evaluate_pseudo_energy(int begin, int end, string structure, string constraint, int frequency)	{
	int i;
	int pseudo_energy = 0;
	
	for(i = begin; i <= end; ++ i)	{
		//cout << "**************" << endl;
		//cout << i << ":	" << reactivity[i] << endl;
		if(structure[i] == '(' || structure[i] == ')')	{
			pseudo_energy += pseudo_energy_pair_constrained(i, constraint[i], frequency);
		}	else if(structure[i] == '.')	{
			pseudo_energy += pseudo_energy_unpair_constrained(i, constraint[i], frequency);
		}
	}
	return pseudo_energy;
}

int shape_reactivity::evaluate_pseudo_energy_exclude(int begin, int end, int begin_exclude, int end_exclude, 
	string structure, string constraint, int frequency)	{

	int i;
	int pseudo_energy = 0;
	for(i = begin; i < begin_exclude; ++ i)	{
		if(structure[i] == '.')	{
			pseudo_energy += pseudo_energy_unpair_constrained(i, constraint[i], frequency);
		}	else if(structure[i] == '(' || structure[i] == ')')	{
			pseudo_energy += pseudo_energy_pair_constrained(i, constraint[i], frequency);
		}
	}
	
	for(i = end_exclude + 1; i <= end; ++ i)	{
		if(structure[i] == '.')	{
			pseudo_energy += pseudo_energy_unpair_constrained(i, constraint[i], frequency);
		}	else if(structure[i] == '(' || structure[i] == ')')	{
			pseudo_energy += pseudo_energy_pair_constrained(i, constraint[i], frequency);
		}
	}
	return pseudo_energy;
}


void shape_reactivity::adjust_unconstrained_energy(string inferred_str, int &energy)	{
	int i;
	for(i = 0; i < inferred_str.length(); ++ i)	{
		if(inferred_str[i] == '(' || inferred_str[i] == ')')	{
			energy -= pseudo_energy_pair(i);
			energy += (int) (100.0 * intercept);
		}
	}
	return;
}

int shape_reactivity::optimistic_pseudo_energy_stack(vector<int> stack)	{
	return (int) 100.0 * intercept * 2 * stack[2];
}

int shape_reactivity::optimistic_pseudo_energy_pair(void)	{
	return (int) 100.0 * intercept;
}

int shape_reactivity::optimistic_pseudo_energy_structure(vector<int> stack_in_str, vector< vector<int> > stacks)	{
	int i;
	int pseudo_energy = 0;
	for(i = 0; i < stack_in_str.size(); ++ i)	{
		pseudo_energy += optimistic_pseudo_energy_stack(stacks[stack_in_str[i]]);
	}
	return pseudo_energy;
}
