#include "./inc/single_fold_stack.h"
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include <stack>
#include <vector>

using namespace std;

/* used for calling related functions without allocating memory */
single_fold_stack::single_fold_stack(string input_sequence)	{
	sequence = input_sequence;
	return;
}

single_fold_stack::single_fold_stack(string input_sequence, string input_SHAPE_file, int min_stack_length, int min_stack_hydro_bonds)	{
		
	sequence = input_sequence;
	structure = string(sequence.length(), 'x');
		
	int i, j;
	stack_hash = vector< vector<int> > (sequence.length());
	for(i = 0; i < sequence.length(); ++ i)	{
		stack_hash[i] = vector<int> (sequence.length());
		for(j = 0; j < sequence.length(); ++ j)	{
			stack_hash[i][j] = -1;
		}
	} 
	
	predict_stacks(min_stack_length, min_stack_hydro_bonds);
	sort_stacks();
	build_stacks();
	
	shape_reactivity SHAPE_constraint(sequence, input_SHAPE_file);
	reactivity = SHAPE_constraint.reactivity;
				
	energy_matrix = vector<int> (stacks.size());
	traceback_matrix = vector<int> (stacks.size());
	traceback_matrix_type = vector<int> (stacks.size());
	
	energy_matrix_multi = vector< vector<int> > (stacks.size());
	traceback_matrix_multi = vector< vector<int> > (stacks.size());
	traceback_matrix_multi_type = vector< vector<int> > (stacks.size());
		
	for(i = 0; i < stacks.size(); ++ i)	{
		energy_matrix_multi[i] = vector<int> (stacks.size());
		traceback_matrix_multi[i] = vector<int> (stacks.size());
		traceback_matrix_multi_type[i] = vector<int> (stacks.size());
	}
	
	return;
}

single_fold_stack::~single_fold_stack(void)	{
	return;
}

int single_fold_stack::num_hydro_bonds(char a, char b)	{
	
	a = tolower(a);
	b = tolower(b);
	
	if(a == 'a' && b == 'u' || a == 'u' && b == 'a')	{
		return 2;
	}	else if(a == 'c' && b == 'g' || a == 'g' && b == 'c')	{
		return 3;
	}	else if(a == 'g' && b == 'u' || a == 'u' && b == 'g')	{
		return 1;
	}	else	{
		return -LARGENUM;
	}
}

int single_fold_stack::predict_stacks(int min_length, int min_hydro_bonds)	{
	int i, j, k;
	
	energy_parameters energy_model(sequence);
	
	vector< vector<int> > sampled_pairs(sequence.length());
	for(i = 0; i < sequence.length(); ++ i)	{
		sampled_pairs[i] = vector<int> (sequence.length());
		for(j = 0; j < sequence.length(); ++ j)	{
			sampled_pairs[i][j] = 0;
		}
	}
	
	for(i = 0; i < sequence.length() - 4; ++ i)	{
		for(j = i + 4; j < sequence.length(); ++ j)	{
			int stack_hydro_bonds = num_hydro_bonds(sequence[i], sequence[j]);
			if(stack_hydro_bonds < 0 || sampled_pairs[i][j] == 1) continue;
			k = 1;
			/* find continuous stackings */
			while(j - i - 2 * k - 1 >= 3)	{
				int p_hydro = num_hydro_bonds(sequence[i + k], sequence[j - k]);
				if(p_hydro > 0)	stack_hydro_bonds += p_hydro;
				else	break;
				++ k;
			}
			/* removing terminal GU pairs */
			int l = 1;
			/*******************************
			while(l <= k)	{
				int p_hydro = num_hydro_bonds(sequence[i + k - l], sequence[j - k + l]);
				if(p_hydro == 1)	stack_hydro_bonds -= p_hydro;
				else	break;
				++ l;
			}
			*******************************/
			
			if(stack_hydro_bonds >= min_hydro_bonds && k - l + 1 >= min_length)	{
				stacks.resize(stacks.size() + 1);
				stacks[stacks.size() - 1] = vector<int> (4);
				stacks[stacks.size() - 1][0] = i;
				stacks[stacks.size() - 1][1] = j;
				stack_hash[i][j] = stacks.size() - 1;
				stacks[stacks.size() - 1][2] = k - l + 1;
				int m;
				int energy = 0;
				for(m = k - l - 1; m >= 0; -- m)	{
					energy += energy_model.FE_stack(i + m, j - m);
				}
				stacks[stacks.size() - 1][3] = energy;
				for(m = 0; m < k - l + 1; ++ m)	{
					sampled_pairs[i + m][j - m] = 1;
				} 
				
			}
		}
	}
	
	/* add the pseudo stack to cover the entire structure */
	stacks.resize(stacks.size() + 1);
	stacks[stacks.size() - 1] = vector<int> (4);
	stacks[stacks.size() - 1][0] = 0;
	stacks[stacks.size() - 1][1] = sequence.length() - 1;
	stacks[stacks.size() - 1][2] = 0;
	stacks[stacks.size() - 1][3] = 0;
	return stacks.size();
}


void single_fold_stack::sort_stacks()	{
	
	int i, j;
	for(i = 0; i < stacks.size() - 1; ++ i)	{
		for(j = i + 1; j < stacks.size(); ++ j)	{
			if(stacks[i][1] > stacks[j][1] || (stacks[i][1] == stacks[j][1] && stacks[i][0] < stacks[j][0]))	{
				vector<int> temp(4);
				temp = stacks[i]; stacks[i] = stacks[j]; stacks[j] = temp;
			}
		}
	}
	
	return;
	
}

void single_fold_stack::build_stacks(void)	{
	
	int i, j;
	
	/* building the internal relationship */
	num_internal = vector<int> (stacks.size());
	internal = vector< vector<int> > (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		num_internal[i] = 0;
		internal[i] = vector<int> (stacks.size());
	}
	
	for(i = stacks.size() - 1; i >= 0; -- i )	{
		for(j = i - 1; j >= 0; -- j)	{
			if(stacks[j][0] > stacks[i][0] + stacks[i][2] - 1 && stacks[j][1] < stacks[i][1] - stacks[i][2] + 1 &&
				stacks[j][0] - stacks[i][0] - stacks[i][2] <= 15 && 
				stacks[i][1] - stacks[i][2] - stacks[j][1] <= 15)	{
				internal[i][num_internal[i]] = j;
				++ num_internal[i];
			}
		}
	}
	
	/* building the enclosing relationship */
	num_enclosing = vector<int> (stacks.size());
	enclosing = vector< vector<int> > (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		num_enclosing[i] = 0;
		enclosing[i] = vector<int> (stacks.size());
	}
	
	for(i = stacks.size() - 1; i >= 0; -- i )	{
		for(j = i - 1; j >= 0; -- j)	{
			if(stacks[j][0] > stacks[i][0] + stacks[i][2] - 1 && stacks[j][1] < stacks[i][1] - stacks[i][2] + 1)	{
				enclosing[i][num_enclosing[i]] = j;
				++ num_enclosing[i];
			}
		}
	}
	
	/* building the FP relationship */
	num_FPset = vector<int> (stacks.size());
	FPset = vector< vector<int> > (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		num_FPset[i] = 0;
		FPset[i] = vector<int> (stacks.size());
	}
	for(i = stacks.size() - 1; i >= 0; -- i)	{
		int boundary = stacks[i][0] + stacks[i][2];
		for(j = i - 1; j >= 0; -- j)	{
			if(stacks[j][1] >= boundary && stacks[j][0] >= stacks[i][0] + stacks[i][2] && stacks[j][1] <= stacks[i][1] - stacks[i][2])	{
				if(stacks[j][0] > boundary)	{
					boundary = stacks[j][0];
				}
				FPset[i][num_FPset[i]] = j;
				++ num_FPset[i];
			}
		}
	}
	
	/* building the F relationship */
	num_Fset = vector<int> (stacks.size());
	Fset = vector< vector<int> > (stacks.size());
	for(i = 0; i < stacks.size(); ++ i)	{
		num_Fset[i] = 0;
		Fset[i] = vector<int> (stacks.size());
	}
	for(i = stacks.size() - 1; i >= 0; -- i)	{
		int boundary = 0;
		for(j = i - 1; j >= 0; -- j)	{
			if(stacks[j][1] >= boundary && stacks[j][1] < stacks[i][0])	{
				if(stacks[j][0] > boundary)	{
					boundary = stacks[j][0];
				}
				Fset[i][num_Fset[i]] = j;
				++ num_Fset[i];
			}
		}
	}
	
	/*
	cout << "Printing all stacks" << endl;
	for(i = 0; i < stacks.size(); ++ i)	{
		cout << i << ":	" << stacks[i][0] << "	" << stacks[i][1] << "	" << stacks[i][2] << "	" << stacks[i][3] << endl;
	}
	cout << "End of printing all stacks" << endl;
	*/
	/*
	for(i = 0; i < stacks.size(); ++ i)	{
		cout << i << ":	";
		for(j = 0; j < num_internal[i]; ++ j)	{
			cout << internal[i][j] << "	";
		}
		cout << endl;
	}
	*/
	return;
}

void single_fold_stack::clear_structure(void)	{
	structure = string(sequence.length(), 'x');
	return;
}

void single_fold_stack::fold()	{
	
	int i, j, k;
	energy_parameters energy_model(sequence);

	for(i = 0; i < stacks.size(); ++ i)	{
		
		//cout << "i:	" << i << endl;
		
		if(i == stacks.size() - 1)	{
			energy_model.set_pseudo_stack_energy();
		}
		
		/* compute hairpin loop energy */
		int intermediate_hairpin = stacks[i][3] + energy_model.FE_hairpin(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1);

		/* compute internal loop energy */
		int intermediate_internal = LARGENUM;
		int min_internal;
		if(num_internal[i] > 0)	{
			for(j = num_internal[i] - 1; j >= 0; -- j)	{
				int x = internal[i][j];
				int hold_internal = stacks[i][3] + energy_matrix[x] 
					+ energy_model.FE_loop(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1, stacks[x][1], stacks[x][0]);
					//cout << x << " internal energy:	" << energy_model.FE_loop(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1, stacks[x][1], stacks[x][0]) << endl;
				if(hold_internal < intermediate_internal)	{
					intermediate_internal = hold_internal;
					min_internal = x;
				}
			}	// endfor
		}	// endif

		/* compute multi loop energy */
		int intermediate_multi = LARGENUM;
		int min_multi;
		if(num_enclosing[i] > 0)	{
			for(j = num_enclosing[i] - 1; j >= 0; -- j)	{
				int x = enclosing[i][j];
				int hold_multi = LARGENUM;
				int min_F;
				for(k = num_Fset[x] - 1; k >= 0; -- k)	{
					int u = Fset[x][k];
					if(stacks[u][0] >= stacks[i][0] + stacks[i][2] && energy_matrix_multi[i][u] < hold_multi)	{
						hold_multi = energy_matrix_multi[i][u];
						min_F = u;
					}
				}
				
				/* the region before the stack can form no structure */
				if(hold_multi >= 0)	{
					hold_multi = 0;
					min_F = -1;
				}
				traceback_matrix_multi[i][x] = min_F;
				//cout << "assignment:	" << i << "	" << x << "	" << min_F << endl;
				
				/* the stack itself can adopt no structure */
				if(energy_matrix[x] + energy_model.Multi_new_branch > 0)	{
					traceback_matrix_multi_type[i][x] = 0;
				}	else	{
					hold_multi += energy_matrix[x] + energy_model.Multi_new_branch;
					traceback_matrix_multi_type[i][x] = 1;
				}
				energy_matrix_multi[i][x] = hold_multi;
				
				hold_multi += stacks[i][3] + energy_model.Multi_closing;
				if(hold_multi < intermediate_multi)	{
					intermediate_multi = hold_multi;
					min_multi = x;
				}
			}
		}
		//cout << intermediate_hairpin << "	" << intermediate_internal << "	" << intermediate_multi << endl;

		if(intermediate_hairpin <= intermediate_internal && intermediate_hairpin <= intermediate_multi)	{
			//cout << "best hairpin" << endl;
			energy_matrix[i] = intermediate_hairpin;
			traceback_matrix_type[i] = 0;
		}	else if(intermediate_internal <= intermediate_hairpin && intermediate_internal <= intermediate_multi)	{
			//cout << "best internal" << endl;
			energy_matrix[i] = intermediate_internal;
			traceback_matrix_type[i] = 1;
			traceback_matrix[i] = min_internal;
		}	else if(intermediate_multi <= intermediate_hairpin && intermediate_multi <= intermediate_internal)	{
			//cout << "best multi" << endl;
			energy_matrix[i] = intermediate_multi;
			traceback_matrix_type[i] = -1;
			traceback_matrix[i] = min_multi;
		}
		//cout << energy_matrix[i] << endl;
	}
	fold_energy = energy_matrix[energy_matrix.size() - 1];
	return;
}

void single_fold_stack::fold_SHAPE()	{
	
	int i, j, k;
	energy_parameters energy_model(sequence);
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	
	for(i = 0; i < stacks.size(); ++ i)	{
		
		if(i == stacks.size() - 1)	{
			energy_model.set_pseudo_stack_energy();
		}
		
		//cout << "i:	" << i << endl;
		//cout << "stack contribution:	" << SHAPE_constraint.pseudo_energy_stack(stacks[i]) << endl;
		/* compute hairpin loop energy */
		int intermediate_hairpin = stacks[i][3] + SHAPE_constraint.pseudo_energy_stack(stacks[i]) 
			+ energy_model.FE_hairpin(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1);

		/* compute internal loop energy */
		int intermediate_internal = LARGENUM;
		int min_internal;
		if(num_internal[i] > 0)	{
			for(j = num_internal[i] - 1; j >= 0; -- j)	{
				int x = internal[i][j];
				int hold_internal = stacks[i][3] + SHAPE_constraint.pseudo_energy_stack(stacks[i]) + energy_matrix[x] 
					+ energy_model.FE_loop(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1, stacks[x][1], stacks[x][0]);
				if(hold_internal < intermediate_internal)	{
					intermediate_internal = hold_internal;
					min_internal = x;
				}
			}	// endfor
		}	// endif

		/* compute multi loop energy */
		int intermediate_multi = LARGENUM;
		int min_multi;
		if(num_enclosing[i] > 0)	{
			for(j = num_enclosing[i] - 1; j >= 0; -- j)	{
				int x = enclosing[i][j];
				int hold_multi = LARGENUM;
				int min_F;
				for(k = num_Fset[x] - 1; k >= 0; -- k)	{
					int u = Fset[x][k];
					if(stacks[u][0] >= stacks[i][0] + stacks[i][2] && energy_matrix_multi[i][u] < hold_multi)	{
						hold_multi = energy_matrix_multi[i][u];
						min_F = u;
					}
				}
				/* the region before the stack can form no structure */
				if(hold_multi >= 0)	{
					hold_multi = 0;
					min_F = -1;
				}
				traceback_matrix_multi[i][x] = min_F;
				//cout << "assignment:	" << i << "	" << x << "	" << min_F << endl;
				
				/* the stack itself can adopt no structure */
				if(energy_matrix[x] + energy_model.Multi_new_branch > 0)	{
					traceback_matrix_multi_type[i][x] = 0;
				}	else	{
					hold_multi += energy_matrix[x] + energy_model.Multi_new_branch;
					traceback_matrix_multi_type[i][x] = 1;
				}
				energy_matrix_multi[i][x] = hold_multi;
				
				hold_multi += stacks[i][3] + SHAPE_constraint.pseudo_energy_stack(stacks[i]) + energy_model.Multi_closing;
				if(hold_multi < intermediate_multi)	{
					intermediate_multi = hold_multi;
					min_multi = x;
				}
			}
		}

		if(intermediate_hairpin <= intermediate_internal && intermediate_hairpin <= intermediate_multi)	{
			//cout << "best hairpin" << endl;
			energy_matrix[i] = intermediate_hairpin;
			traceback_matrix_type[i] = 0;
		}	else if(intermediate_internal <= intermediate_hairpin && intermediate_internal <= intermediate_multi)	{
			//cout << "best internal" << endl;
			energy_matrix[i] = intermediate_internal;
			traceback_matrix_type[i] = 1;
			traceback_matrix[i] = min_internal;
		}	else if(intermediate_multi <= intermediate_hairpin && intermediate_multi <= intermediate_internal)	{
			//cout << "best multi" << endl;
			energy_matrix[i] = intermediate_multi;
			traceback_matrix_type[i] = -1;
			traceback_matrix[i] = min_multi;
		}
		//cout << energy_matrix[i] << endl;
	}
	fold_energy = energy_matrix[energy_matrix.size() - 1];
	return;
}

void single_fold_stack::fold_SHAPE_constrained(string constrained_str, int frequency)	{
	
	int i, j, k;
	energy_parameters energy_model(sequence);
	shape_reactivity SHAPE_constraint(sequence, reactivity);
	
	for(i = 0; i < stacks.size(); ++ i)	{
		
		if(i == stacks.size() - 1)	{
			energy_model.set_pseudo_stack_energy();
		}
		
		//cout << "i:	" << i << endl;
		//cout << "stack contribution:	" << SHAPE_constraint.pseudo_energy_stack(stacks[i]) << endl;
		/* compute hairpin loop energy */
		int intermediate_hairpin = stacks[i][3] + SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i])
			+ SHAPE_constraint.pseudo_energy_stack_constrained(stacks[i], constrained_str, frequency)
			+ energy_model.FE_hairpin(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1);

		/* compute internal loop energy */
		int intermediate_internal = LARGENUM;
		int min_internal;
		if(num_internal[i] > 0)	{
			for(j = num_internal[i] - 1; j >= 0; -- j)	{
				int x = internal[i][j];
				int hold_internal = stacks[i][3] + energy_matrix[x] + SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i])
					+ SHAPE_constraint.pseudo_energy_stack_constrained(stacks[i], constrained_str, frequency) 
					+ energy_model.FE_loop(stacks[i][0] + stacks[i][2] - 1, stacks[i][1] - stacks[i][2] + 1, stacks[x][1], stacks[x][0]);
				if(hold_internal < intermediate_internal)	{
					intermediate_internal = hold_internal;
					min_internal = x;
				}
			}	// endfor
		}	// endif

		/* compute multi loop energy */
		int intermediate_multi = LARGENUM;
		int min_multi;
		if(num_enclosing[i] > 0)	{
			for(j = num_enclosing[i] - 1; j >= 0; -- j)	{
				int x = enclosing[i][j];
				int hold_multi = LARGENUM;
				int min_F;
				for(k = num_Fset[x] - 1; k >= 0; -- k)	{
					int u = Fset[x][k];
					if(stacks[u][0] >= stacks[i][0] + stacks[i][2] && energy_matrix_multi[i][u] < hold_multi)	{
						hold_multi = energy_matrix_multi[i][u];
						min_F = u;
					}
				}
				/* the region before the stack can form no structure */
				if(hold_multi >= 0)	{
					hold_multi = 0;
					min_F = -1;
				}
				traceback_matrix_multi[i][x] = min_F;
				//cout << "assignment:	" << i << "	" << x << "	" << min_F << endl;
				
				/* the stack itself can adopt no structure */
				int mstack_energy = energy_matrix[x] + energy_model.Multi_new_branch + SHAPE_constraint.optimistic_pseudo_energy_stack(stacks[i])
					+ SHAPE_constraint.pseudo_energy_stack_constrained(stacks[i], constrained_str, frequency);
				if(mstack_energy > 0)	{
					traceback_matrix_multi_type[i][x] = 0;
				}	else	{
					hold_multi += mstack_energy;
					traceback_matrix_multi_type[i][x] = 1;
				}
				energy_matrix_multi[i][x] = hold_multi;
				
				hold_multi += stacks[i][3] + SHAPE_constraint.pseudo_energy_stack(stacks[i]) + energy_model.Multi_closing;
				if(hold_multi < intermediate_multi)	{
					intermediate_multi = hold_multi;
					min_multi = x;
				}
			}
		}

		if(intermediate_hairpin <= intermediate_internal && intermediate_hairpin <= intermediate_multi)	{
			//cout << "best hairpin" << endl;
			energy_matrix[i] = intermediate_hairpin;
			traceback_matrix_type[i] = 0;
		}	else if(intermediate_internal <= intermediate_hairpin && intermediate_internal <= intermediate_multi)	{
			//cout << "best internal" << endl;
			energy_matrix[i] = intermediate_internal;
			traceback_matrix_type[i] = 1;
			traceback_matrix[i] = min_internal;
		}	else if(intermediate_multi <= intermediate_hairpin && intermediate_multi <= intermediate_internal)	{
			//cout << "best multi" << endl;
			energy_matrix[i] = intermediate_multi;
			traceback_matrix_type[i] = -1;
			traceback_matrix[i] = min_multi;
		}
		//cout << energy_matrix[i] << endl;
	}
	fold_energy = energy_matrix[energy_matrix.size() - 1];
	return;
}

void single_fold_stack::append_stack(vector<int> stack, string &structure)	{
	string left(stack[2], '(');
	string right(stack[2], ')');
	structure.replace(stack[0], stack[2], left);
	structure.replace(stack[1] - stack[2] + 1, stack[2], right);
	return;
}

void single_fold_stack::traceback_stack(int i)	{
	
	if(i < 0 || i >= stacks.size())	{
		cout << "Error: traceback called for a stack that is out of range" << endl;
		return;
	}
	
	structure.replace(stacks[i][0], stacks[i][1] - stacks[i][0] + 1, string(stacks[i][1] - stacks[i][0] + 1, '.'));
	append_stack(stacks[i], structure);
	if(traceback_matrix_type[i] == 1)	{
		//cout << "internal loop" << endl;
		traceback_stack(traceback_matrix[i]);
	}	else if(traceback_matrix_type[i] == -1)	{
		//cout << "multi loop" << endl;
		int m = traceback_matrix[i];
		while(m >= 0)	{
			//cout << "m:	" << m << " in " << i << endl;
			if(traceback_matrix_multi_type[i][m] == 1)	traceback_stack(m);
			m = traceback_matrix_multi[i][m];
			//cout << "new m:	" << m << endl;
		}
	}
	return;
}

void single_fold_stack::traceback_multi(int i, int m)	{
	
	if(i < 0 || i >= stacks.size())	{
		cout << "Error: traceback called for a stack that is out of range" << endl;
		return;
	}	
	
	if(m >= i || stacks[m][0] < stacks[i][0] + stacks[i][2] || stacks[m][1] > stacks[i][1] - stacks[i][2])	{
		cout << "Error: traceback called for a multi loop case out of range" << endl;
		return;
	}
	
	structure.replace(stacks[i][0] + stacks[i][2], stacks[m][1] - stacks[i][0] - stacks[i][2] + 1, 
		string(stacks[m][1] - stacks[i][0] - stacks[i][2] + 1, '.'));
	while(m >= 0)	{
		//cout << "m:	" << m << " in " << i << endl;
		if(traceback_matrix_multi_type[i][m] == 1)	traceback_stack(m);
		m = traceback_matrix_multi[i][m];
		//cout << "new m:	" << m << endl;
	}
	
	return;
}

int single_fold_stack::evaluate_energy_stack(vector<int> stack_in_str)	{
	energy_parameters energy_model(sequence);
	int energy = 0;
	/* sorting the stack ids in stack_in_str */
	int i, j;
	for(i = 0; i < stack_in_str.size() - 1; ++ i)	{
		for(j = i + 1; j < stack_in_str.size(); ++ j)	{
			if(stack_in_str[i] > stack_in_str[j])	{
				int temp = i;
				i = j;
				j = temp;
			}
		}
	}
	/* traverse each stack and compute corresponding energy */
	vector<int> outmost_stacks;
	for(i = 0; i < stack_in_str.size(); ++ i)	{
		int m = stack_in_str[i];
		for(j = 0; j < outmost_stacks.size(); ++ j)	{
			int n = outmost_stacks[j]; 
			if(stacks[m][0] + stacks[m][2] - 1 < stacks[n][0])	{
				break;
			}
		}
		
		if(j == outmost_stacks.size())	{
			/* adding a hairpin loop energy */
			energy += energy_model.FE_hairpin(stacks[m][0] + stacks[m][2] - 1, stacks[m][1] - stacks[m][2] + 1);
			energy += stacks[m][3];	
			/* the new stack covers no one, add a new multi loop */
			outmost_stacks.resize(outmost_stacks.size() + 1);
			outmost_stacks[outmost_stacks.size() - 1] = m;
		}	else if(j == outmost_stacks.size() - 1)	{
			/* adding internal/bulge loop energy */
			int x = outmost_stacks[outmost_stacks.size() - 1];
			energy += energy_model.FE_loop(stacks[m][0] + stacks[m][2] - 1, stacks[m][1] - stacks[m][2] + 1, stacks[x][1], stacks[x][0]);
			energy += stacks[m][3];
			/* the new stack covers the last one, replace the last loop */
			outmost_stacks[outmost_stacks.size() - 1] = m;
		}	else if(outmost_stacks.size() - j > 1)	{
			/* adding multi loop energy */
			energy += energy_model.Multi_closing;
			energy += energy_model.Multi_new_branch * (outmost_stacks.size() - j);
			energy += stacks[m][3];
			/* the new stack covers several other stacks, record the last */
			outmost_stacks.resize(j + 1);
			outmost_stacks[outmost_stacks.size() - 1] = m;
		}
	}
	return energy;
}
